﻿'
' HINWEIS für Bilder mit Alpha-Werten:
' Bei der Darstellung von Bildern mit Alpha-Werten ungleich 255(=keine Transparenz) oder
' ungleich 0(=vollständige Transparenz), kommt es zu Anzeigefehlern: 
' Da der Hintergrund des Buttons nicht neugezeichnet, sondern immer weiter überschrieben 
' wird, werden die halbtransparenten Bilder übereinander gelagert, sodass die
' Transparenz mit jedem Zeichenvorgang abnimmt.
' Dies kann verhindert werden, i dem der Button-Hintergrund manuell neu gezeichnet wird.
' Das Ereignis ButtonForceBGRefresh informiert darüber, wenn das Neuzeichnen geschehen muss.
' Sollte das Skinbutton direkt auf einem Formular liegen, reicht es, den Befehl 
' "Me.Refresh" auf das Forumlar auszuführen (immer, wenn das Ereignis eintritt).
'
Public Class SkinButton
    Inherits System.Windows.Forms.UserControl

    Public Const Version = "1.0.2"

    Public StandardPicPostfix As String = "_normal"
    Public DisabledPicPostfix As String = "_disabled"
    Public HoveredPicPostfix As String = "_hovered"
    Public FocusedPicPostfix As String = "_focused"
    Public PressedPicPostfix As String = "_pressed"

    Private Const WS_EX_TRANSPARENT As Int32 = &H20
    Private Const KeyPressedWaitTime = 60 ' Milliseconds, siehe auch: Event SkinButton_KeyPress()

    Public Enum eStates
        Standard
        Disabled
        Hovered
        Focused
        Pressed
    End Enum
    Public Enum eSpecialEffects
        None
        PositionChange
    End Enum
    Dim myState As eStates = eStates.Standard
    Dim mySkinPictureFFile As String = ""
    Dim mySkinPictureFile As String = ""
    Dim myTColor As Color = Color.Red
    Dim myUseTColor As Boolean = True
    Dim myButtonText As String = ""
    Dim myStandardPic As Bitmap
    Dim myDisabledPic As Bitmap
    Dim myHoveredPic As Bitmap
    Dim myFocusedPic As Bitmap
    Dim myPressedPic As Bitmap
    Dim FilenameEnding As String = ""
    Dim myIsHovered As Boolean = False
    Dim mySupportsFocusing As Boolean = False
    Dim myUseKeysForClick As Boolean = True
    Dim mySkinPictureWidth As Integer = 0
    Dim mySkinPictureHeight As Integer = 0
    Dim mySpecialEffect As eSpecialEffects = eSpecialEffects.None
    Dim myButtonTextFontname As String = "Microsoft Sans Serif"
    Dim myButtonTextFontsize As Single = 8
    Dim myButtonTextFontStyle As System.Drawing.FontStyle = FontStyle.Regular
    Dim myButtonTextColor As Color = Color.Black
    Dim BTFont As New Font(myButtonTextFontname, myButtonTextFontsize)
    Dim BTSF As New StringFormat(StringFormatFlags.MeasureTrailingSpaces)
    Dim BTBrush As New SolidBrush(myButtonTextColor)

    Public Sub New()
        ' Dieser Aufruf ist für den Windows Form-Designer erforderlich.
        InitializeComponent()
        ' Fügen Sie Initialisierungen nach dem InitializeComponent()-Aufruf hinzu.

        SetStyle(ControlStyles.SupportsTransparentBackColor, True)
        'SetStyle(ControlStyles.UserPaint, True)
        'SetStyle(ControlStyles.AllPaintingInWmPaint, True)
        'SetStyle(ControlStyles.OptimizedDoubleBuffer, True)

        Me.UpdateStyles()
        Me.BackColor = Color.Transparent

        BTSF.Alignment = StringAlignment.Center
        BTSF.LineAlignment = StringAlignment.Center
    End Sub

    Protected Overrides ReadOnly Property CreateParams() As CreateParams
        Get
            Dim cp As CreateParams = MyBase.CreateParams
            cp.ExStyle = cp.ExStyle Or WS_EX_TRANSPARENT
            Return cp
        End Get
    End Property

    Protected Overrides Sub OnPaintBackground(ByVal e As PaintEventArgs)
        ' (Optionaler Code?)
    End Sub

    Public Delegate Sub ButtonClickDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event ButtonClick As ButtonClickDelegate

    Public Delegate Sub ButtonMouseClickDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event ButtonMouseClick As ButtonMouseClickDelegate

    Public Delegate Sub ButtonDblClickDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event ButtonDblClick As ButtonDblClickDelegate

    Public Delegate Sub ButtonMouseDblClickDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event ButtonMouseDblClick As ButtonMouseDblClickDelegate

    Public Delegate Sub ButtonKeyPressDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyPressEventArgs)
    Public Shadows Event ButtonKeyPress As ButtonKeyPressDelegate

    Public Delegate Sub ButtonForceBGRefreshDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event ButtonForceBGRefresh As ButtonForceBGRefreshDelegate

    Private Sub SkinButton_GotFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.GotFocus
        If mySupportsFocusing Then
            If myState = eStates.Standard Or myState = eStates.Hovered Then
                myState = eStates.Focused
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
            End If
        End If
    End Sub

    Private Sub SkinButton_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LostFocus
        If mySupportsFocusing Then
            If myState = eStates.Focused Then
                If myIsHovered Then
                    myState = eStates.Hovered
                Else
                    myState = eStates.Standard
                End If
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
            End If
        End If
    End Sub

    Private Sub SkinButton_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        If Not myState = eStates.Disabled Then
            If IsPixelClickable(e.X, e.Y) Then
                myState = eStates.Pressed
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
            End If
        End If
    End Sub

    Private Sub UserControl1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        Dim gr As Graphics = e.Graphics
        Dim bm As Bitmap

        If mySkinPictureFile = "" Then Exit Sub

        Select Case State
            Case eStates.Standard
                bm = myStandardPic
            Case eStates.Disabled
                bm = myDisabledPic
            Case eStates.Hovered
                bm = myHoveredPic
            Case eStates.Focused
                bm = myFocusedPic
            Case eStates.Pressed
                bm = myPressedPic
        End Select
        gr.DrawImage(bm, 0, 0)
        mySkinPictureWidth = bm.Height
        mySkinPictureHeight = bm.Width

        If Not myButtonText = "" Then
            Dim BTRect As Rectangle
            If mySpecialEffect = eSpecialEffects.PositionChange And myState = eStates.Pressed Then
                BTRect = New Rectangle(1, 1, Me.Width, Me.Height)
            Else
                BTRect = New Rectangle(0, 0, Me.Width, Me.Height)
            End If

            gr.DrawString(myButtonText, BTFont, BTBrush, BTRect, BTSF)
        End If
    End Sub

    Private Sub SkinButton_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        If IsPixelClickable(e.X, e.Y) And Not myState = eStates.Disabled Then
            RaiseEvent ButtonClick(Me, e)
            RaiseEvent ButtonMouseClick(Me, e)
        End If
    End Sub

    <System.ComponentModel.Browsable(True)> _
    Public ReadOnly Property SkinPictureWidth() As Integer
        Get
            Return mySkinPictureWidth
        End Get
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public ReadOnly Property SkinPictureHeight() As Integer
        Get
            Return mySkinPictureHeight
        End Get
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property SpecialEffect() As eSpecialEffects
        Get
            Return mySpecialEffect
        End Get
        Set(ByVal value As eSpecialEffects)
            mySpecialEffect = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ButtonTextColor() As Color
        Get
            Return myButtonTextColor
        End Get
        Set(ByVal value As Color)
            myButtonTextColor = value
            BTBrush.Color = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ButtonTextFontname() As String
        Get
            Return myButtonTextFontname
        End Get
        Set(ByVal value As String)
            myButtonTextFontname = value

            BTFont = New Font(myButtonTextFontname, myButtonTextFontsize, myButtonTextFontStyle)
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ButtonTextFontsize() As Single
        Get
            Return myButtonTextFontsize
        End Get
        Set(ByVal value As Single)
            Try
                BTFont = New Font(myButtonTextFontname, myButtonTextFontsize, myButtonTextFontStyle)
            Catch ex As Exception
                Err.Raise(513, "SkinButton", "Error: This buttontext-fontsize is invalid: " + value.ToString)
            End Try

            myButtonTextFontsize = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ButtonTextFontStyle() As System.Drawing.FontStyle
        Get
            Return myButtonTextFontStyle
        End Get
        Set(ByVal value As System.Drawing.FontStyle)
            myButtonTextFontStyle = value

            BTFont = New Font(myButtonTextFontname, myButtonTextFontsize, myButtonTextFontStyle)
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property SupportsFocusing() As Boolean
        Get
            Return mySupportsFocusing
        End Get
        Set(ByVal value As Boolean)
            mySupportsFocusing = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property UseKeysForClick() As Boolean
        Get
            Return myUseKeysForClick
        End Get
        Set(ByVal value As Boolean)
            myUseKeysForClick = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property State() As eStates
        Get
            Return myState
        End Get
        Set(ByVal value As eStates)
            myState = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property SkinPictureFile() As String
        Get
            Return mySkinPictureFile
        End Get
        Set(ByVal value As String)
            mySkinPictureFile = value
            If Not mySkinPictureFile = "" Then
                Dim Pos As Integer
                Pos = InStrRev(mySkinPictureFile, ".")
                If Pos = 0 Then
                    FilenameEnding = ""
                Else
                    FilenameEnding = Mid(mySkinPictureFile, Pos)
                    mySkinPictureFFile = Microsoft.VisualBasic.Left(mySkinPictureFile, Pos - 1)
                End If
                Try
                    myStandardPic = New Bitmap(mySkinPictureFFile + StandardPicPostfix + FilenameEnding)
                    If myUseTColor Then myStandardPic.MakeTransparent(myTColor)
                    mySkinPictureWidth = myStandardPic.Width
                    mySkinPictureHeight = myStandardPic.Height
                Catch ex As Exception
                    mySkinPictureFile = ""
                    mySkinPictureWidth = 0
                    mySkinPictureHeight = 0
                    Err.Raise(513, "SkinButton", "Error: Could not load SkinButton-picture: " + mySkinPictureFFile + StandardPicPostfix + FilenameEnding)
                    Exit Property
                End Try
                Try
                    myFocusedPic = New Bitmap(mySkinPictureFFile + FocusedPicPostfix + FilenameEnding)
                    If myUseTColor Then myFocusedPic.MakeTransparent(myTColor)
                Catch ex As Exception
                    Try
                        myHoveredPic = New Bitmap(mySkinPictureFFile + HoveredPicPostfix + FilenameEnding)
                        If myUseTColor Then myHoveredPic.MakeTransparent(myTColor)
                    Catch ex2 As Exception
                        myHoveredPic = myStandardPic
                    End Try
                    myFocusedPic = myHoveredPic
                End Try
                Try
                    myPressedPic = New Bitmap(mySkinPictureFFile + PressedPicPostfix + FilenameEnding)
                    If myUseTColor Then myPressedPic.MakeTransparent(myTColor)
                Catch ex As Exception
                    myPressedPic = myStandardPic
                End Try
                Try
                    myDisabledPic = New Bitmap(mySkinPictureFFile + DisabledPicPostfix + FilenameEnding)
                    If myUseTColor Then myDisabledPic.MakeTransparent(myTColor)
                Catch ex As Exception
                    myDisabledPic = myStandardPic
                End Try
            Else
                myStandardPic.Dispose()
                myFocusedPic.Dispose()
                myHoveredPic.Dispose()
                myPressedPic.Dispose()
                myDisabledPic.Dispose()
                mySkinPictureWidth = 0
                mySkinPictureHeight = 0
            End If
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TransparencyColor() As Drawing.Color
        Get
            Return myTColor
        End Get
        Set(ByVal value As Drawing.Color)
            myTColor = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property UseTransparencyColor() As Boolean
        Get
            Return myUseTColor
        End Get
        Set(ByVal value As Boolean)
            myUseTColor = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ButtonText() As String
        Get
            Return myButtonText
        End Get
        Set(ByVal value As String)
            myButtonText = value
        End Set
    End Property

    Private Sub SkinButton_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        If Not myState = eStates.Disabled Then
            If mySupportsFocusing Then
                myState = eStates.Focused
            Else
                If myIsHovered Then
                    myState = eStates.Hovered
                Else
                    myState = eStates.Standard
                End If
            End If
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End If
    End Sub

    Private Function IsPixelClickable(ByVal X As Integer, ByVal Y As Integer) As Boolean
        Dim bm As Bitmap

        Select Case myState
            Case eStates.Standard
                bm = myStandardPic
            Case eStates.Hovered
                bm = myHoveredPic
            Case eStates.Focused
                bm = myFocusedPic
            Case eStates.Pressed
                bm = myPressedPic
            Case Else
                Exit Function
        End Select

        If X < 0 Or X > bm.Width - 1 Then Return False
        If Y < 0 Or Y > bm.Height - 1 Then Return False
        Dim PointColor As Color = bm.GetPixel(X, Y)

        Return Not (PointColor.A = 0) ' A = Alphawert, 0 = 100% transparent
    End Function

    Private Sub SkinButton_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseLeave
        myIsHovered = False
        If myState = eStates.Hovered Then
            myState = eStates.Standard
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End If
    End Sub

    Private Sub SkinButton_MouseEnter(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseEnter
        myIsHovered = True
        If myState = eStates.Standard Then
            myState = eStates.Hovered
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End If
    End Sub

    Private Sub SkinButton_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDoubleClick
        If IsPixelClickable(e.X, e.Y) And Not myState = eStates.Disabled Then
            RaiseEvent ButtonDblClick(Me, e)
            RaiseEvent ButtonMouseDblClick(Me, e)
        End If
    End Sub

    Private Sub SkinButton_KeyPress(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyPressEventArgs) Handles Me.KeyPress
        If myState = eStates.Disabled Then Exit Sub

        RaiseEvent ButtonKeyPress(Me, e)
        If UseKeysForClick = False Then Exit Sub
        If e.KeyChar = Microsoft.VisualBasic.ChrW(Keys.Return) Or e.KeyChar = Microsoft.VisualBasic.ChrW(Keys.Space) Then
            myState = eStates.Pressed
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
            RaiseEvent ButtonClick(Me, New System.Windows.Forms.MouseEventArgs(Windows.Forms.MouseButtons.None, 0, 0, 0, 0))

            Dim SaveTime As Date = Now
            Do
                Application.DoEvents()
            Loop Until (Now - SaveTime).Milliseconds > KeyPressedWaitTime

            If mySupportsFocusing Then
                myState = eStates.Focused
            Else
                If myIsHovered Then
                    myState = eStates.Hovered
                Else
                    myState = eStates.Standard
                End If
            End If
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End If
    End Sub

    Private Sub InvalidateMe()
        Me.Invalidate()
        RaiseEvent ButtonForceBGRefresh(Me, New System.EventArgs())
    End Sub
End Class

