﻿Imports System.Windows.Forms
Imports System.Drawing

Public Class SkinLabel
    Inherits System.Windows.Forms.PictureBox

    Public Const Version = "0.7"

    Private WithEvents myLabel As Windows.Forms.Label
    Private myPicture As String = ""
    Private myAutoSizeContainer As Boolean = True
    Private myMultilined As Boolean = False
    Private myFont As System.Drawing.Font
    Private myTextFontname As String = "Microsoft Sans Serif"
    Private myTextFontsize As Single = 8
    Private myTextFontStyle As System.Drawing.FontStyle = FontStyle.Regular
    Private components As System.ComponentModel.IContainer
    Private myAutoTicker As Boolean = False
    Private myTickerTime As Date
    Private TickerTimer As Windows.Forms.Timer
    Private TickerX As Single = 0
    Private TickerY As Single = 0
    Private myTickerSpeed As Single = 0.2

    Public Sub New()
        SetStyle(ControlStyles.SupportsTransparentBackColor, True)

        Me.UpdateStyles()
        Me.BackColor = Color.Transparent

        myFont = New System.Drawing.Font(myTextFontname, myTextFontsize, myTextFontStyle)

        myLabel = New Windows.Forms.Label
        myLabel.Location = New Point(0, 0)
        myLabel.AutoSize = True
        myLabel.Text = ""
        myLabel.Font = myFont
        Controls.Add(myLabel)

        TickerTimer = New Windows.Forms.Timer
        TickerTimer.Interval = 10
        AddHandler TickerTimer.Tick, AddressOf TickerTimer_Tick
    End Sub

    Public Delegate Sub TextClickDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event TextClick As TextClickDelegate

    Public Delegate Sub TextMouseClickDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event TextMouseClick As TextMouseClickDelegate

    Public Delegate Sub TextDblClickDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event TextDblClick As TextDblClickDelegate

    Public Delegate Sub TextMouseDblClickDelegate(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event TextMouseDblClick As TextMouseDblClickDelegate

    Public Delegate Sub HoveringDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event Hovering As HoveringDelegate

    Public Delegate Sub ForceBGRefreshDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event ForceBGRefresh As ForceBGRefreshDelegate

    Private Sub TickerTimer_Tick()
        If myLabel.Text > "" Then
            If TickerX + myLabel.Width > Me.Width Then
                TickerX -= myTickerSpeed
                myLabel.Location = New Point(TickerX, TickerY)
                If TickerX + myLabel.Width <= Me.Width Then
                    myTickerTime = Now
                End If
            Else
                If (Now - myTickerTime).Seconds >= 1 Then
                    TickerX = 0
                    myLabel.Location = New Point(TickerX, TickerY)
                End If
            End If
        End If
    End Sub

    Private Sub SkinLabel_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        RaiseEvent TextClick(Me, e)
        RaiseEvent TextMouseClick(Me, e)
    End Sub

    <System.ComponentModel.Browsable(True)> _
    Public Property TextColor() As Color
        Get
            Return myLabel.ForeColor
        End Get
        Set(ByVal value As Color)
            myLabel.ForeColor = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TextFontname() As String
        Get
            Return myTextFontname
        End Get
        Set(ByVal value As String)
            myFont = New System.Drawing.Font(value, myTextFontsize, myTextFontStyle)
            myLabel.Font = myFont
            myTextFontname = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TextFontsize() As Single
        Get
            Return myTextFontsize
        End Get
        Set(ByVal value As Single)
            Try
                myFont = New System.Drawing.Font(myTextFontname, value, myTextFontStyle)
                myLabel.Font = myFont
                myTextFontsize = value
            Catch ex As Exception
                Err.Raise(513, "SkinLabel", "Error: This buttontext-fontsize is invalid: " + value.ToString)
            End Try
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TextFontStyle() As System.Drawing.FontStyle
        Get
            Return myTextFontStyle
        End Get
        Set(ByVal value As System.Drawing.FontStyle)
            myFont = New System.Drawing.Font(myTextFontname, myTextFontsize, value)
            myLabel.Font = myFont
            myTextFontStyle = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Overrides Property Text() As String
        Get
            Return myLabel.Text
        End Get
        Set(ByVal value As String)
            myLabel.Text = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TextPosition() As Point
        Get
            Return myLabel.Location
        End Get
        Set(ByVal value As Point)
            myLabel.Location = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property AutoTicker() As Boolean
        Get
            Return myAutoTicker
        End Get
        Set(ByVal value As Boolean)
            myAutoTicker = value
            TickerTimer.Enabled = value
            TickerX = myLabel.Location.X
            TickerY = myLabel.Location.Y
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ContainerImage() As String
        Get
            Return myPicture
        End Get
        Set(ByVal value As String)
            If value = "" Then
                Me.BackgroundImage = Nothing
            Else
                Try
                    Me.BackgroundImage = New Bitmap(value)
                    myPicture = value
                    If myAutoSizeContainer Then
                        Me.Width = Me.BackgroundImage.Width
                        Me.Height = Me.BackgroundImage.Height
                        If myMultilined Then
                            myLabel.AutoSize = False ' eig. redundant
                            myLabel.SetBounds(0, 0, Me.Width, Me.Height)
                        Else
                            myLabel.AutoSize = True ' eig. redundant
                            myLabel.Location = New Point(0, Me.Height / 2 - myLabel.Height / 2)
                        End If
                    End If
                Catch ex As Exception
                    Err.Raise(513, "SkinLabel", "Error: Could not load picture: " + value)
                End Try
            End If
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property AutoSizeContainer() As Boolean
        Get
            Return myAutoSizeContainer
        End Get
        Set(ByVal value As Boolean)
            myAutoSizeContainer = value

            If myAutoSizeContainer And myPicture > "" Then
                Me.Width = Me.BackgroundImage.Width
                Me.Height = Me.BackgroundImage.Height
                If myMultilined Then
                    myLabel.AutoSize = False ' eig. redundant
                    myLabel.SetBounds(0, 0, Me.Width, Me.Height)
                Else
                    myLabel.AutoSize = True ' eig. redundant
                    myLabel.Location = New Point(0, Me.Height / 2 - myLabel.Height / 2)
                End If
            End If
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property Multilined() As Boolean
        Get
            Return myMultilined
        End Get
        Set(ByVal value As Boolean)
            myMultilined = value
            If myMultilined Then
                myLabel.AutoSize = False
                myLabel.SetBounds(0, 0, Me.Width, Me.Height)
            Else
                myLabel.AutoSize = True
                myLabel.Location = New Point(0, Me.Height / 2 - myLabel.Height / 2)
            End If
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TickerSpeed() As Single
        Get
            Return myTickerSpeed
        End Get
        Set(ByVal value As Single)
            If value < 0 Then
                Err.Raise(513, "SkinLabel", "Invalid ticker speed value: " + value.ToString)
            Else
                myTickerSpeed = value
            End If
        End Set
    End Property

    ' Angabe in Pixeln
    Public Function GetTextWidth() As Integer
        Return myLabel.Width
    End Function

    Private Sub SkinLabel_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDoubleClick
        RaiseEvent TextDblClick(Me, e)
        RaiseEvent TextMouseDblClick(Me, e)
    End Sub

    Private Sub InitializeComponent()
        CType(Me, System.ComponentModel.ISupportInitialize).BeginInit()
        Me.SuspendLayout()
        '
        'SkinLabel
        '
        CType(Me, System.ComponentModel.ISupportInitialize).EndInit()
        Me.ResumeLayout(False)

    End Sub

    Private Sub myLabel_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles myLabel.MouseClick
        RaiseEvent TextClick(Me, e)
        RaiseEvent TextMouseClick(Me, e)
    End Sub

    Private Sub myLabel_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles myLabel.MouseDoubleClick
        RaiseEvent TextDblClick(Me, e)
        RaiseEvent TextMouseDblClick(Me, e)
    End Sub

    Private Sub SkinLabel_MouseHover(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseHover
        RaiseEvent Hovering(Me, e)
    End Sub

    Private Sub myLabel_MouseHover(ByVal sender As Object, ByVal e As System.EventArgs) Handles myLabel.MouseHover
        RaiseEvent Hovering(Me, e)
    End Sub
End Class
