﻿
Public Class SkinPicture
    Inherits System.Windows.Forms.UserControl

    Public Version = "1.0.0"

    Private Const WS_EX_TRANSPARENT As Int32 = &H20

    Private mySkinPictureFile As String
    Private myTColor As Color
    Private myUseTColor As Boolean
    Private myBitmap As Bitmap
    Private mySkinPictureWidth As Integer = 0
    Private mySkinPictureHeight As Integer = 0
    Private myAlphaValue As Byte

    Public Delegate Sub PictureClickDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event PictureClick As PictureClickDelegate

    Public Delegate Sub ForceBGRefreshDelegate(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event ForceBGRefresh As ForceBGRefreshDelegate

    Public Sub New()
        ' Dieser Aufruf ist für den Windows Form-Designer erforderlich.
        InitializeComponent()
        ' Fügen Sie Initialisierungen nach dem InitializeComponent()-Aufruf hinzu.

        SetStyle(ControlStyles.SupportsTransparentBackColor, True)

        Me.UpdateStyles()
        Me.BackColor = Color.Transparent
    End Sub

    Protected Overrides ReadOnly Property CreateParams() As CreateParams
        Get
            Dim cp As CreateParams = MyBase.CreateParams
            cp.ExStyle = cp.ExStyle Or WS_EX_TRANSPARENT
            Return cp
        End Get
    End Property

    Protected Overrides Sub OnPaintBackground(ByVal e As PaintEventArgs)
        ' (Optionaler Code?)
    End Sub

    Private Sub SkinPicture_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        If IsPixelClickable(e.X, e.Y) Then
            RaiseEvent PictureClick(Me, e)
        End If
    End Sub

    Private Sub SkinPicture_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        Dim gr As Graphics = e.Graphics

        If Not mySkinPictureFile = "" Then
            gr.DrawImage(myBitmap, 0, 0)
        End If
    End Sub

    <System.ComponentModel.Browsable(True)> _
    Public Property SkinPictureFile() As String
        Get
            Return mySkinPictureFile
        End Get
        Set(ByVal value As String)
            If Not value = "" Then
                Try
                    myBitmap = New Bitmap(value)
                    mySkinPictureWidth = myBitmap.Width
                    mySkinPictureHeight = myBitmap.Height
                    If myUseTColor Then myBitmap.MakeTransparent(myTColor)
                    mySkinPictureFile = value
                Catch ex As Exception
                    mySkinPictureWidth = 0
                    mySkinPictureHeight = 0
                    mySkinPictureFile = ""
                    Err.Raise(513, "SkinPicture", "Could not load SkinPicutre: " + value)
                    Exit Property
                End Try
            Else
                myBitmap.Dispose()
                mySkinPictureWidth = 0
                mySkinPictureHeight = 0
            End If
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property AlphaValue() As Byte
        Get
            Return myAlphaValue
        End Get
        Set(ByVal value As Byte)
            myAlphaValue = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TransparencyColor() As Color
        Get
            Return myTColor
        End Get
        Set(ByVal value As Color)
            myTColor = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property UseTransparencyColor() As Boolean
        Get
            Return myUseTColor
        End Get
        Set(ByVal value As Boolean)
            myUseTColor = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von UserControl1_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public ReadOnly Property SkinPictureWidth() As Integer
        Get
            Return mySkinPictureWidth
        End Get
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public ReadOnly Property SkinPictureHeight() As Integer
        Get
            Return mySkinPictureHeight
        End Get
    End Property

    Private Sub InvalidateMe()
        Me.Invalidate()
        RaiseEvent ForceBGRefresh(Me, New System.EventArgs())
    End Sub

    ' Gibt true zurück, wenn ein angegebener Pixel sichtbar (und somit anklickbar) ist
    Private Function IsPixelClickable(ByVal X As Integer, ByVal Y As Integer) As Boolean
        If mySkinPictureFile = "" Then Return False

        If X < 0 Or X > myBitmap.Width - 1 Then Return False
        If Y < 0 Or Y > myBitmap.Height - 1 Then Return False
        Dim PointColor As Color = myBitmap.GetPixel(X, Y)

        Return Not (PointColor.A = 0) ' A = Alphawert, 0 = 100% transparent
    End Function
End Class
