﻿Imports System.Windows.Forms
Imports System.Drawing

Public Class SkinProgressBar
    Inherits PictureBox

    Public Const Version = "0.9"
    Public Enum eMode
        eVertical
        eHorizontal
    End Enum

    Private Const WS_EX_TRANSPARENT As Int32 = &H20
    Private myMode As eMode = eMode.eHorizontal
    Private myPictureForeground As String
    Private myPictureBar As String
    Private myPictureBackground As String
    Private myPictureButtonNormal As String
    Private myPictureButtonHovered As String
    Private myBMForeground As Bitmap
    Private myBMBar As Bitmap
    Private myBMBackground As Bitmap
    Private myBMButtonNormal As Bitmap
    Private myBMButtonHovered As Bitmap
    Private myProgress As Single = 0 ' 0-100 (Prozent)
    Private myValue As Single = 0 ' 0-100 (Prozent)
    Private myVisualStep As Single = -1 ' 0-100 (Prozent), -1 = keine Verwendung
    Private myTColor As Drawing.Color = Color.Red
    Private myUseTColor As Boolean = True
    Private myIsChangeable As Boolean = True
    Private myMinPosition As New Point(-1, -1) ' X/Y für 0% Fortschritt, -1 = auto
    Private myMaxPosition As New Point(-1, -1) ' X/Y für 100% Fortschritt, -1 = auto
    Private myReversed As Boolean = False ' bei horizontal: Fortschritt nach links anstatt rechts, bei vertikal: Fortschritt nach unten anstatt nach oben.
    Private mySmoothMoveEnabled As Boolean = False  ' Bei gedrückter Maustaste kann der Fortschritt sutfenlos verändert werden.
    Private IsMouseDown As Boolean = False
    Private SButton As New SkinButton.SkinButton
    Private Test As New Windows.Forms.Button
    Private myHasButton As Boolean = False
    Private myShowButton As Boolean = True
    Private myButtonHovered As Boolean = False
    Private myButtonPosition As New Point(0, 0)

    Public Sub New()
        'SetStyle(ControlStyles.SupportsTransparentBackColor, True)

        'Me.UpdateStyles()
        'Me.BackColor = Color.Transparent
    End Sub

    Protected Overrides ReadOnly Property CreateParams() As CreateParams
        Get
            Dim cp As CreateParams = MyBase.CreateParams
            cp.ExStyle = cp.ExStyle Or WS_EX_TRANSPARENT
            Return cp
        End Get
    End Property

    Protected Overrides Sub OnPaintBackground(ByVal e As PaintEventArgs)
        ' (Optionaler Code?)
    End Sub

    Public Delegate Sub DelegateForceBGRefresh(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Shadows Event ForceBGRefresh As DelegateForceBGRefresh

    Public Delegate Sub DelegateChange(ByVal sender As Object, ByVal e As System.EventArgs, ByVal NewValue As Single)
    Public Shadows Event Change As DelegateChange
    Public Shadows Event ChangeSmoothed As DelegateChange

    Public Delegate Sub DelegatePBClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs)
    Public Shadows Event PBClick As DelegatePBClick
    Public Shadows Event PBMouseClick As DelegatePBClick

    Private Sub SkinProgressBar_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        RaiseEvent PBClick(sender, e)
        RaiseEvent PBMouseClick(sender, e)
    End Sub

    Private Sub SkinProgressBar_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        If Not e.Button = Windows.Forms.MouseButtons.Left Then Exit Sub

        IsMouseDown = True

        Dim AddX As Integer = 0, AddY As Integer = 0
        CalcButtonAdds(AddX, AddY)
        CalculateChange(New Point(e.X + AddX, e.Y + AddY))
    End Sub

    Private Sub SkinProgressBar_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove
        If mySmoothMoveEnabled And IsMouseDown Then
            Dim AddX As Integer = 0, AddY As Integer = 0
            CalcButtonAdds(AddX, AddY)
            CalculateChange(New Point(e.X + AddX, e.Y + AddY), True)
        End If
    End Sub

    Private Sub SkinProgressBar_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        If Not e.Button = Windows.Forms.MouseButtons.Left Then Exit Sub

        IsMouseDown = False

        If mySmoothMoveEnabled Then
            Dim AddX As Integer = 0, AddY As Integer = 0
            CalcButtonAdds(AddX, AddY)
            CalculateChange(New Point(e.X + AddX, e.Y + AddY))
        End If
    End Sub

    ' Kleine Hilfsmethode; dient dazu, einige Zeilen Code zu sparen.
    Private Sub CalcButtonAdds(ByRef AddX As Integer, ByRef AddY As Integer)
        If myHasButton And myShowButton Then
            If myButtonHovered Then
                If myMode = eMode.eHorizontal Then AddX = myBMButtonHovered.Width / 2 Else AddY = myBMButtonHovered.Height / 2
            Else
                If myMode = eMode.eHorizontal Then AddX = myBMButtonNormal.Width / 2 Else AddY = myBMButtonNormal.Height / 2
            End If
        End If
    End Sub

    Private Sub CalculateChange(ByVal Pos As Point, Optional ByVal RaiseEventChangeSmoothed As Boolean = False)
        If myMode = eMode.eHorizontal Then
            If Pos.X <= myMinPosition.X Then
                myProgress = 0
            Else
                If Pos.X >= myMaxPosition.X Then
                    myProgress = 100
                Else
                    myProgress = (Pos.X - myMinPosition.X) / (myMaxPosition.X - myMinPosition.X) * 100
                End If
            End If
            If myReversed Then myProgress = 100 - myProgress
        Else
            If Pos.Y <= myMinPosition.Y Then
                myProgress = 0
            Else
                If Pos.Y >= myMaxPosition.Y Then
                    myProgress = 100
                Else
                    myProgress = (Pos.Y - myMinPosition.Y) / (myMaxPosition.Y - myMinPosition.Y) * 100
                End If
            End If
            If myReversed = False Then myProgress = 100 - myProgress
        End If
        If IsMouseDown = False Or SmoothMoveEnabled = False Then
            Value = myProgress
            RaiseEvent Change(Me, New System.EventArgs(), myProgress)
        End If
        If RaiseEventChangeSmoothed Then RaiseEvent ChangeSmoothed(Me, New System.EventArgs(), myProgress)
        If myVisualStep > -1 Then myProgress = Int(myProgress / myVisualStep) * myVisualStep
        InvalidateMe()
    End Sub

    Private Sub SkinProgressBar_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles Me.Paint
        Dim gr As Graphics = e.Graphics
        Dim bi As Image = Nothing ' bi = button image

        If myPictureBackground > "" Then
            gr.DrawImage(myBMBackground, 0, 0) ' Hintergrund
        End If
        If myPictureBar > "" Then
            If myMode = eMode.eHorizontal Then
                If myReversed Then
                    gr.DrawImage(myBMBar, myMaxPosition.X - (myMaxPosition.X - myMinPosition.X) * myProgress / 100, myMinPosition.Y + (myMaxPosition.Y - myMinPosition.Y) * myProgress / 100) ' nach links
                Else
                    gr.DrawImage(myBMBar, myMinPosition.X + (myMaxPosition.X - myMinPosition.X) * myProgress / 100 - myBMBar.Width, myMinPosition.Y + (myMaxPosition.Y - myMinPosition.Y) * myProgress / 100) ' nach rechts
                End If
            Else
                If myReversed Then
                    gr.DrawImage(myBMBar, myMinPosition.X + (myMaxPosition.X - myMinPosition.X) * myProgress / 100, myMinPosition.Y + (myMaxPosition.Y - myMinPosition.Y) * myProgress / 100 - myBMBar.Height) ' nach unten 
                Else
                    gr.DrawImage(myBMBar, myMinPosition.X + (myMaxPosition.X - myMinPosition.X) * myProgress / 100, myMinPosition.Y - (myMaxPosition.Y - myMinPosition.Y) * myProgress / 100 + myBMBar.Height) ' nach oben
                End If
            End If
            If myPictureForeground > "" Then
                gr.DrawImage(myBMForeground, 0, 0) ' Vordergrund
            End If
        End If
        If myHasButton And myShowButton Then
            If myButtonHovered Then
                bi = myBMButtonHovered
            Else
                bi = myBMButtonNormal
            End If
            If myMode = eMode.eHorizontal Then
                If myReversed Then
                    ' nach links
                Else
                    myButtonPosition = New Point(MinPosition.X + (myMaxPosition.X - myMinPosition.X) * myProgress / 100 - bi.Width, myMinPosition.Y + (myMaxPosition.Y - myMinPosition.Y) * myProgress / 100) ' nach rechts
                End If
            Else
                If myReversed Then
                    ' nach unten 
                Else
                    ' nach oben
                End If
            End If
            gr.DrawImage(bi, myButtonPosition.X, myButtonPosition.Y)
        End If
    End Sub

    <System.ComponentModel.Browsable(True)> _
    Public Property Mode() As eMode
        Get
            Return myMode
        End Get
        Set(ByVal value As eMode)
            myMode = value
            Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property PictureForeground() As String
        Get
            Return myPictureForeground
        End Get
        Set(ByVal value As String)
            Try
                If Not value = "" Then
                    myBMForeground = New Bitmap(value)
                    If myUseTColor Then myBMForeground.MakeTransparent(myTColor)
                    Me.Width = myBMForeground.Width
                    Me.Height = myBMForeground.Height
                End If
                myPictureForeground = value
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
            Catch ex As Exception
                Err.Raise(513, "SkinProgressBar", "Could not load picture: " + value)
            End Try
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property PictureBar() As String
        Get
            Return myPictureBar
        End Get
        Set(ByVal value As String)
            Try
                If Not value = "" Then
                    myBMBar = New Bitmap(value)

                    If myUseTColor Then myBMBar.MakeTransparent(myTColor)
                End If
                myPictureBar = value
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
            Catch ex As Exception
                Err.Raise(513, "SkinProgressBar", "Could not load picture: " + value)
            End Try
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property PictureBackground() As String
        Get
            Return myPictureBackground
        End Get
        Set(ByVal value As String)
            Try
                If Not value = "" Then
                    myBMBackground = New Bitmap(value)
                    If myUseTColor Then myBMBackground.MakeTransparent(myTColor)
                    Me.Width = myBMBackground.Width
                    Me.Height = myBMBackground.Height
                End If
                myPictureBackground = value
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
            Catch ex As Exception
                Err.Raise(513, "SkinProgressBar", "Could not load picture: " + value)
            End Try
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public ReadOnly Property HasButton() As Boolean
        Get
            Return myHasButton
        End Get
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property TransparencyColor() As Drawing.Color
        Get
            Return myTColor
        End Get
        Set(ByVal value As Drawing.Color)
            myTColor = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property UseTransparencyColor() As Boolean
        Get
            Return myUseTColor
        End Get
        Set(ByVal value As Boolean)
            myUseTColor = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property MinPosition() As Point
        Get
            Return myMinPosition
        End Get
        Set(ByVal value As Point)
            myMinPosition = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property MaxPosition() As Point
        Get
            Return myMaxPosition
        End Get
        Set(ByVal value As Point)
            myMaxPosition = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property Value() As Single
        Get
            Return myValue
        End Get
        Set(ByVal value As Single)
            If value < 0 Or value > 100 Then
                Err.Raise(513, "SkinProgressBar", "Error: Invalid value. Value must between 0 and 100. Your (invalid) value: " + value.ToString)
            Else
                myValue = value
                If IsMouseDown = False Or SmoothMoveEnabled = False Then
                    myProgress = myValue
                    If myVisualStep > -1 Then myProgress = Int(myProgress / myVisualStep) * myVisualStep
                    Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
                End If
            End If
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property Reversed() As Boolean
        Get
            Return myReversed
        End Get
        Set(ByVal value As Boolean)
            myReversed = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property IsChangeable() As Boolean
        Get
            Return myIsChangeable
        End Get
        Set(ByVal value As Boolean)
            myIsChangeable = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property SmoothMoveEnabled() As Boolean
        Get
            Return mySmoothMoveEnabled
        End Get
        Set(ByVal value As Boolean)
            mySmoothMoveEnabled = value
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property VisualStep() As Single
        Get
            Return myVisualStep
        End Get
        Set(ByVal value As Single)
            myVisualStep = value
            If myVisualStep < 1 Or myVisualStep > 100 Then myVisualStep = -1
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property ShowButton() As Boolean
        Get
            Return myShowButton
        End Get
        Set(ByVal value As Boolean)
            myShowButton = value
            InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property PictureButtonNormal() As String
        Get
            Return myPictureButtonNormal
        End Get
        Set(ByVal value As String)
            Try
                If Not value = "" Then
                    myBMButtonNormal = New Bitmap(value)
                    If myUseTColor Then myBMButtonNormal.MakeTransparent(myTColor)
                End If
                myPictureButtonNormal = value
                myHasButton = (value <> "")
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
            Catch ex As Exception
                Err.Raise(513, "SkinProgressBar", "Could not load picture: " + value)
            End Try
        End Set
    End Property

    <System.ComponentModel.Browsable(True)> _
    Public Property PictureButtonHovered() As String
        Get
            Return myPictureButtonHovered
        End Get
        Set(ByVal value As String)
            Try
                If Not value = "" Then
                    myBMButtonHovered = New Bitmap(value)
                    If myUseTColor Then myBMButtonHovered.MakeTransparent(myTColor)
                End If
                myPictureButtonHovered = value
                Me.InvalidateMe() ' erzwingt das Neuzeichnen des UserControls, also den Aufruf von SkinProgressBar_Paint()
            Catch ex As Exception
                Err.Raise(513, "SkinProgressBar", "Could not load picture: " + value)
            End Try
        End Set
    End Property

    Private Sub InvalidateMe()
        RaiseEvent ForceBGRefresh(Me, New System.EventArgs())
        Me.Invalidate()
    End Sub

    Private Sub InitializeComponent()
        CType(Me, System.ComponentModel.ISupportInitialize).BeginInit()
        Me.SuspendLayout()
        '
        'SkinProgressBar
        '
        CType(Me, System.ComponentModel.ISupportInitialize).EndInit()
        Me.ResumeLayout(False)
    End Sub
End Class
