﻿'
' acore = Audia Core Engine
' Dieses Modul stellt die Grundfunktionen von Audia zur Verfügung.
'
Module acore
    Public Const Version = "7.7.0"
    Public Const VersionBuild = 7700
    Public Const IsVoid = -1 ' Globale Konstante, an vielen Stellen verwendet. Bedeutung: leer/unbelegt/unbenutzt
    Public Const SoundFilenameExtensions = "WAV|MP3|OGG|WMA|MID|S3M|RAW|MOD" ' Bekannte, unterstützte Dateiendungen für Sounddateien
    Public Const PlaylistFilenameExtensions = "M3U|PLS|WPL|XSPF|TXT" ' Bekannte, unterstützte Dateiendungen für Playlisten
    Public Const PlaylistDialogFilter = "Alle Playlisten | *.M3U;*.PLS;*.WPL;*.XSPF;*.TXT"
    Public Const MediaDialogFilter = "Alle Dateien | *.*"
    Public Const DefaultPlaylistName = "last.m3u" ' Diese Playlist wird beim Start des Players geladen und beim Beenden gespeichert
    Public Const SettingsFile = "settings.ini" ' Diese Datei speichert die Einstellungen
    Public Const LogFile = "Log.txt" ' Dateiname ohne Pfadangabe der Protokolldatei (der Pfad ist das Audia-Arbeitsverzeichnis); leerer Eintrag = keine Protokollierung
    Public Const MaxLogFileSize = 100000 ' legt fest, ab welcher Dateigröße in Bytes die Log-Datei (festgelegt durch LogFile) gelöscht (= geleert) wird
    Public Const LibraryFilename = "MediaLibrary.lib" ' Dateiname der Datei, in der die Medienbibliothek gespeichert wird
    Public Const FavoritsFilename = "Favorits.lib" ' Dateiname der Datei, in der die Favoriten gespeichert werden
    Public Const VolumeStep = 10 ' Prozentwert. Gibt an, wie viel % die Laustärke pro Schritt erhöht/verringert wird.
    Public Const PosStep = 10 ' Sekunden. GIbt na, wie viele Sekunden die Abspielposition pro Schritt verändert wird.
    Public Const PortableSettingsFilename = "audia_portable_settings.aps" ' Dateiname.
    Public Const MaxItemCountInGotoMediumMenu = 50 ' Wie viele Einträge darf das GotoMedium-Menü max. haben? (Bei zu vielen Einträgen dauert es viel zu lange, die Items hinzuzufügen, außerdme wir passt das Menü nicht mehr auf den Bildschirm)
    Public Const LinkMusicDownload = "http://www.musicload.de/" ' Stellt einen Link zu einer Seite für Bezahldownloads bereit
    Public Const LinkLYRDB = "http://www.lyrdb.com/" '  Stellt einen Link zur Website von LYRDB.com bereit
    Public Const DefaultLanguage = "German" ' Gibt das Schlüsselwort für die (nicht lokalisierte) Standartsprache an (auf Englisch)
    Public Const LangDir = "localized" ' Ordnername, in dem die (sprach)lokalisierten Dateien liegen
    Public Const HelpDir = "help" ' Ordnername, in dem die Hilfedateien liegen
    Public Const AppUpdateURL = "http://www.myinx.de/data/audia/update.auf" ' URL zu einer Datei mit Update-Informationen bezüglich des Programms

    Public Enum eMediumTypes
        eUndefined ' unbekannter Typ
        eError ' Fehler beim Abspielen (evtl. nicht unterstütztes Format)
        eSoundfile
        eNetstream
        eVideo
    End Enum

    Public Structure sMedium
        Dim FModSound As FMOD.Sound
        Dim Filename As String ' Dateiname oder URL. Bsp.: "C:\Test\Mozart - Konzert 1.mp3"
        Dim FileTitle As String ' Bsp.: "Mozart - Konzert 1", also der Dateiname ohne Pfadangabe und Endung
        Dim Title As String ' Titel des Mediums, z.B. Songtitel
        Dim Artist As String ' Name des Interpreten/Künstlers/Band
        Dim MediumID As Integer
        Dim MediumType As eMediumTypes
        Dim Stars As Integer ' Bewertung (1-5 "Sterne")
        Dim PlayLen As Single ' Abspiellänge in Sekunden
        Dim HasPlayLen As Boolean ' Für NetStreams. Manche Streams haben eine Länge, z.B. wenn es sich um eine hochgeladene MP3-Datei handelt.
        Dim IsPlayLenTested As Boolean ' Wurde getestet, ob "HasPlayLen=true" stimmt? (Relevant nur für NetSteams)

        Public Function GetBestTitle() As String
            If Title = "" Then
                Return FileTitle
            Else
                If Artist = "" Then
                    Return Title
                Else
                    Return Artist + " - " + Title
                End If
            End If
        End Function
    End Structure
    Public Media As List(Of sMedium) ' Speichert die Medien(bibliothek)

    Public Structure sPlaylist
        Dim colMediaIDs As List(Of Integer)
        Dim CurMediumIndex As Integer ' Gibt den Index des colMedia-Items zurück, dass das akt. Medium referenziert
        Dim CurMediumID As Integer
        Dim IsForwarding As Boolean ' Gibt an, ob gerade 8versucht) wird, ein Medium der virutllen Playliste abzuspielen
    End Structure
    Public Playlist As sPlaylist ' Gewährt Zugriff auf die "virtuelle" Playliste

    Public Structure sFavorit
        Dim Title As String
        Dim Target As String
    End Structure
    Public Favorits As List(Of sFavorit)

    Public Structure sPortable
        Dim Drive As String
        Dim Name As String
        Dim Enabled As Boolean ' Wird der DAatenträger von Audia verwaltet?
        Dim Use As Boolean ' Ist der Datenträger im Menü zur Benutzung ausgewählt?
    End Structure
    Public Portables As List(Of sPortable) ' Speichert die momentan am PC angeschlossenen portablen Geräte (Flash-Speicher)

    Public AppPath As String
    Public UserVolume As Single = 50 ' 0-100%, Lautstärke der Tonausgabe
    Public DefaultQuickplayURL As String = "http://scfire-ntc-aa07.stream.aol.com:80/stream/1003"
    Public RepeatPlaylist As Boolean = False ' Wenn true, wird die virtuelle Playliste in einer Endlosschleife abgespielt
    Public RepeatMedium As Boolean = False ' Wenn true, wird das akt. Medium wiederholt.
    Public frmAudiaStartupLocation As Point ' Bildschirmkoordinaten 
    Public LogServiceStarted As Boolean = False
    Public RandomMode As Boolean = False ' Zufallswiedergabe aktiviert?
    Public Rnd As Random ' Wird bei Start() initialisiert
    Public LockMainTimer As Boolean = False
    Public IntroModeDuration As Single = 10 ' Wie lange dauern die Intros beim Intromodus? Angabe in Sekunden
    Public IntroMode As Boolean = False ' Intromodus aktiviert?
    Public EnableConsole As Boolean = False ' Kommando-Konsole aktivieren?
    Public WindowOpacity As Single = 1 ' Transparenzwert zwischen 0 & 1
    Public AlwaysTopMost As Boolean ' Immer im Vordergrund?
    Public GlobalTicks As Long = 0 ' "Globale Ticks"; bei jedem Intervall des HauptKontrollTimers wird diese Variable um einen Tick erhöht. Bei "Integer" würde es nach > 1 Jahr zu einem Überlauf kommen.
    Public Language As String ' Englischer Name der aktuellen Sprache, z.B. "German"
    Public RealAppUpdateURL As String = "" ' Speichert den URL für  den Download des "echten" Programmupdates
    Public AllowAppUpdateCheck As Boolean = True ' Legt fest, ob as Programm bei seinem Start via Internet nach einem neuen Update suchen darf.

    Private WithEvents MainTimer As Timer
    Private WithEvents LowLevelTimer As Timer

    ' Startmethode.
    'Public Sub Main()
    '    Application.EnableVisualStyles() ' XP/Vista-Optik aktivieren
    '    acore.Start()
    '    Application.Run(frmAudia)
    'End Sub

    ' Startet den Audia-Kern.
    ' AUFRUF NOTWENDIG!
    Public Sub Start()
        AppPath = Application.StartupPath ' Reihenfolge: Sollte als allererstes stehen.
        Log("Start", Loc("Beginn des Startvorgangs.")) ' Startvorgang protokollieren

        ' Soundengine laden:
        afi.Start()

        ' Initialisierungen:
        Media = New List(Of sMedium)
        Portables = New List(Of sPortable)
        Favorits = New List(Of sFavorit)

        Playlist = New sPlaylist
        Playlist.CurMediumIndex = IsVoid ' IsVoid = unbelegt
        Playlist.CurMediumID = IsVoid ' IsVoid = unbelegt
        Playlist.colMediaIDs = New List(Of Integer)

        Rnd = New Random(Now.Millisecond) ' Globalen Zufallszahlengenerator initialisieren

        ' Medienbibliothek laden:
        LoadLibrary()

        ' Default-Playliste laden: (Reihenfolge: Muss vor dem Laden der Settings stehen, weil in den Settings evtl. eine Playlistpos gesetzt wird!)
        LoadPlaylist(AppPath + "\" + DefaultPlaylistName)

        ' Einstellungen laden:
        Log("Start", Loc("Lade Einstellungen"))
        LoadSettings()

        ' Favoriten laden:
        LoadFavorits()
        UpdateFavoritsMenu()

        ' GUI updaten:
        UpdateGUI(True)

        ' Timer starten:
        MainTimer = New Timer()
        MainTimer.Interval = 10
        MainTimer.Enabled = True
        LowLevelTimer = New Timer()
        LowLevelTimer.Interval = 999
        LowLevelTimer.Enabled = True

        ' Globale Tastenabfrage vorbereiten:
        SetGlobalKeys()

        If My.Application.CommandLineArgs.Count > 0 Then
            Dim FileList(My.Application.CommandLineArgs.Count - 1) As String
            For i As Integer = 0 To My.Application.CommandLineArgs.Count - 1
                FileList(i) = My.Application.CommandLineArgs(i)
            Next
            QuickplayFilelist(FileList)
        End If

        Log("Start", Loc("Ende des Startvorgangs")) ' Ende des Startvorgangs protokollieren

        AppUpdateCheck() ' Auf neue Updates prüfen. Dies geschieht im Hintergrund und verzögert den Programmstart nicht.
    End Sub

    ' Beendet den Audia Kern.
    ' AUFRUF NOTWENDIG!
    Public Sub Shutdown()
        MainTimer.Enabled = False

        SaveLibrary()
        SaveFavorits()
        SaveSettings()
        SavePlaylist(AppPath + "\" + DefaultPlaylistName)

        afi.Shutdown()

        Log("Shutdown", Loc("Programm beendet.")) ' Ende protokollieren
        End
    End Sub

    ' Lädt die Medienbibliothek aus einer Datei in die virtuelle Medienbibliothek
    Public Function LoadLibrary() As Boolean
        Log("LoadLibrary", Loc("Lade Medienbibliothek (Beginn)")) ' Beginn des Ladevorgangs protokollieren

        Media.Clear() ' virtuelle  Medienbibliothek leeren

        Try
            Dim sr As New IO.StreamReader(AppPath + "\" + LibraryFilename)
            Dim Medium As sMedium = Nothing
            Dim Line As String

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                If Not Line.StartsWith("'") Or Line = "" Then
                    If Line.StartsWith("file=") Then
                        If Not IsNothing(Medium.Filename) Then Media.Add(Medium)
                        Medium = New sMedium ' Bei jedem Vorkommen einer Zeile des Schemas "file=Dateiname" wird ein neue Insatz von sMedium erstellt. Die reienfolge der Zeilen in der Datei ist daher sehr wichtig.
                        Medium.MediumID = Media.Count
                        Medium.Filename = Mid(Line, 6)
                        Medium.FileTitle = GetFileTitle(Medium.Filename)
                        Medium.HasPlayLen = True ' Das wird auch für Netstreams angenommen
                    End If
                    If Line.StartsWith("len=") Then Medium.PlayLen = Val(Mid(Line, 5)) ' Abspiellänge (Zeit)
                    If Line.StartsWith("type=") Then Medium.MediumType = Val(Mid(Line, 6))
                    If Line.StartsWith("title=") Then Medium.Title = Mid(Line, 7)
                    If Line.StartsWith("artist=") Then Medium.Artist = Mid(Line, 8) ' Interpret
                    If Line.StartsWith("stars=") Then Medium.Stars = Val(Mid(Line, 7)) ' Bewertung
                End If
            Loop
            If Not IsNothing(Medium.Filename) Then Media.Add(Medium)

            sr.Close()
        Catch ex As Exception
            Log("LoadLibrary", Loc("Medienbibliothk konnte nicht korrekt ausgelesen werden."))
        End Try

        Log("LoadLibrary", Loc("Lade Medienbibliothek (Abgeschlossen)")) ' Ende des Ladevorgangs protokollieren
    End Function

    ' Speichert die virtuelle Medienbibliothek in eine Datei
    Public Function SaveLibrary() As Boolean
        Dim myFileExists As Boolean

        Log("SaveLibrary", Loc("Speichere Medienbibliothek (Beginn)")) ' Beginn des Speichervorgangs protokollieren

        Try
            Dim sw As New IO.StreamWriter(AppPath + "\" + LibraryFilename + ".tmp") ' Es wird aus Sicherheitsgründen zunächst mit einer temporären Datei gearbeitet. Sollte es zu Fehlern beim Speichern der Medienbib. kommen, bleibt die eigentliche Datei intakt.

            sw.WriteLine("' " + Application.ProductName + " Media Librabry")
            sw.WriteLine("'")

            For Each Medium In Media
                If Not (Medium.MediumType = eMediumTypes.eNetstream Or Medium.MediumType = eMediumTypes.eUndefined Or Medium.MediumType = eMediumTypes.eError) Then
                    If IsPermanent(Medium.Filename) Then ' Medien auf CDs, MP3-Playern etc. werden nicht in der Medienbib. behalten
                        Try
                            myFileExists = My.Computer.FileSystem.FileExists(Medium.Filename)  ' nur existierende Medien werden gespeichert
                        Catch ex As Exception ' Fehler, wenn: Der Name der Datei mit einem umgekehrten Schrägstrich (\) endet (also z.B. bei einem Ordner)
                            Log("SaveLibrary", Loc("Fehler bei Existenzüberprüfung einer Datei. Der Fehler wird ignoriert."))
                            myFileExists = False
                        End Try
                        If myFileExists Then
                            sw.WriteLine("file=" + Medium.Filename)
                            sw.WriteLine("type=" + Val(Medium.MediumType).ToString)
                            sw.WriteLine("len=" + Medium.PlayLen.ToString) ' Abspiellänge
                            sw.WriteLine("stars=" + Medium.Stars.ToString) ' Bewertung
                            sw.WriteLine("title=" + Medium.Title) ' Titel & Interpret werden gespeichert. Das hat Vorteile: Bei Netstreams müsste der Netstream erst geladen werden - was einige Sekunden dauert, um seine Tags auszulesen.
                            sw.WriteLine("artist=" + Medium.Artist) ' Interpret
                        End If
                    End If
                End If
            Next

            sw.Close()

            My.Computer.FileSystem.CopyFile(AppPath + "\" + LibraryFilename + ".tmp", AppPath + "\" + LibraryFilename, True) ' Temporäre Datei nach eigentlicher Datei kopieren
            My.Computer.FileSystem.DeleteFile(AppPath + "\" + LibraryFilename + ".tmp") ' Temp. Datei entfernen
        Catch ex As Exception
            Log("SaveLibrary", Loc("Medienbibliothk konnte nicht gesichert werden."))
        End Try
        Log("SaveLibrary", Loc("Speichere Medienbibliothek (Abgeschlossen)")) ' Ende des Speichervorgangs protokollieren
    End Function

    ' Lädt die Favoriten aus einer Datei
    Public Function LoadFavorits() As Boolean
        Log("LoadFavorits", Loc("Lade Favoriten (Beginn)")) ' Beginn des Ladevorgangs protokollieren

        Favorits.Clear() ' Favoriten leeren

        Try
            Dim sr As New IO.StreamReader(AppPath + "\" + FavoritsFilename)
            Dim Favorit As sFavorit = Nothing
            Dim Line As String

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                If Not Line.StartsWith("'") Or Line = "" Then
                    If Line.StartsWith("target=") Then
                        If Not IsNothing(Favorit.Target) Then Favorits.Add(Favorit)
                        Favorit = New sFavorit ' Bei jedem Vorkommen einer Zeile des Schemas "target=URL" wird ein neue Insatz von sFavorit erstellt. Die reienfolge der Zeilen in der Datei ist daher sehr wichtig.
                        Favorit.Target = Mid(Line, 8) ' Dateiname/URL
                    End If
                    If Line.StartsWith("title=") Then Favorit.Title = Mid(Line, 7) ' Titel
                End If
            Loop
            If Not IsNothing(Favorit.Target) Then Favorits.Add(Favorit)

            sr.Close()
        Catch ex As Exception
            Log("LoadFavorits", Loc("Favoriten konnten nicht korrekt ausgelesen werden."))
        End Try

        Log("LoadFavorits", Loc("Lade Favoriten (Abgeschlossen)")) ' Ende des Ladevorgangs protokollieren
    End Function

    ' Speichert die virtuelle Medienbibliothek in eine Datei
    Public Function SaveFavorits() As Boolean
        Log("SaveFavorits", Loc("Speichere Favoriten (Beginn)")) ' Beginn des Speichervorgangs protokollieren

        Try
            Dim sw As New IO.StreamWriter(AppPath + "\" + FavoritsFilename + ".tmp") ' Es wird aus Sicherheitsgründen zunächst mit einer temporären Datei gearbeitet. Sollte es zu Fehlern beim Speichern der Medienbib. kommen, bleibt die eigentliche Datei intakt.

            sw.WriteLine("' " + Application.ProductName + " Favorits Librabry")
            sw.WriteLine("'")

            For Each Favorit In Favorits
                sw.WriteLine("target=" + Favorit.Target)
                sw.WriteLine("title=" + Favorit.Title)
            Next

            sw.Close()

            My.Computer.FileSystem.CopyFile(AppPath + "\" + FavoritsFilename + ".tmp", AppPath + "\" + FavoritsFilename, True) ' Temporäre Datei nach eigentlicher Datei kopieren
            My.Computer.FileSystem.DeleteFile(AppPath + "\" + FavoritsFilename + ".tmp") ' Temp. Datei entfernen
        Catch ex As Exception
            Log("SaveFavorits", Loc("Favoriten konnten nicht gesichert werden."))
        End Try
        Log("SaveFavorits", Loc("Speichere Favoriten (Abgeschlossen)")) ' Ende des Speichervorgangs protokollieren
    End Function

    ' Lädt ein Medium in die Medienbibliothek
    ' (Die Datei selbst wird nicht geladen!)
    ' Gibt die (neu) zugewiesene MediumID zurück.
    Public Function LoadMedium(ByVal Filename As String) As Long
        Dim FilenameUpper As String = Filename.ToUpper

        For Each myMedium In Media ' Prüfe, ob Medium bereits vorhanden
            If myMedium.Filename.ToUpper = FilenameUpper Then Return myMedium.MediumID ' Falls ja, lediglich die MediumID zurück geben.
        Next

        Dim Medium As New sMedium
        Dim FilenameEnding As String

        Medium.Filename = Filename
        Medium.FileTitle = GetFileTitle(Filename)
        Medium.Title = "" ' wird beim ersten Abspielen ermittelt
        Medium.Artist = "" ' wird beim ersten Abspielen ermittelt
        Medium.Stars = IsVoid ' IsVoid -> noch keine Bewertung erfolgt
        Medium.PlayLen = IsVoid ' IsVoid -> Abspiellänge ist noch unbekannt
        Medium.HasPlayLen = True ' Das wird auch für NetStreams vorerst angenommen.
        ' Typ des Mediums anhand der Dateiendung bestimmen:
        FilenameEnding = GetFilenameExtension(Filename)
        If IsInStringList(SoundFilenameExtensions, FilenameEnding) Then Medium.MediumType = eMediumTypes.eSoundfile
        If IsURL(Filename) Then Medium.MediumType = eMediumTypes.eNetstream
        Medium.MediumID = Media.Count

        Media.Add(Medium)

        Return Medium.MediumID
    End Function

    ' Zeigt einen Dialog zur Dateiauswahl an und spielt die gewählte Datei(en) sofort ab.
    ' Berücksichtig auch Playlisten.
    Public Function Quickplay() As Boolean
        Dim OpenDialog As New OpenFileDialog

        OpenDialog.Filter = MediaDialogFilter
        OpenDialog.Multiselect = True
        OpenDialog.CheckFileExists = True
        OpenDialog.AddExtension = True
        OpenDialog.ShowDialog()

        If OpenDialog.FileNames.Length = 0 Then
            Return False
        Else
            Dim StartPos As Integer = Playlist.colMediaIDs.Count ' Speichert die Position des momentan letzten Playlist-Elements

            Dim Filename As String
            For Each Filename In OpenDialog.FileNames
                If IsPlaylist(Filename) Then
                    ' Sollte die Datei eine Playliste sein, wird sie geladen und die einzelnen Elemente dieser Playliste werden der virtuellen Playliste hinzugefügt.
                    LoadPlaylist(Filename)
                Else
                    AddToPlaylist(LoadMedium(Filename)) ' Zur virtuellen Playliste hinzufügen.
                End If
            Next
            If Playlist.colMediaIDs.Count > StartPos Then ' Diese Bedingung ist notwendig, da ja z.B. eine leere Playliste keine neue Elemente zur virtuellen Playliste hinzufügen würde.
                PlaylistGoTo(StartPos + 1, True) ' +1 , da Index/Pos. des ehemals letzten Elements. Spielt das Medium ab.
            End If
            Return True
        End If
    End Function

    ' ACHTUNG: EXPERIMENTELL! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ' Zeigt eine InputBox zur Eingabe einer URL an (Netstream, Dateiname),
    ' die dann sofort abgespielt wird.
    Public Function QuickplayURL() As Boolean
        Dim URL As String

        URL = InputBox(Loc("Bitte geben Sie eine gültige URL ein, um einen Net-Stream oder eine Datei wiederzugeben.") + vbCr + vbCr + Loc("Hinweis: Playlisten können hiermit nicht abgespielt werden!"), Loc("URL angeben"), DefaultQuickplayURL)
        If URL = "" Then Return False

        If PlayMedium(acore.LoadMedium(URL)) = False Then
            Log("Quickplay Netstream", Loc("Fehler beim Laden eines Netstreams: ") + URL)
            MsgBox(Loc("Ein Fehler ist aufgetreten.") + vbCr + vbCr + Loc("Überprüfen Sie, ob die eingegebene URL korrekt ist und erreicht werden kann. Möglicherweise besteht auch keine Verbindung zum Internert."), vbCritical, Loc("Fehler"))
            Return False
        Else
            DefaultQuickplayURL = URL
            Return True
        End If
    End Function

    ' Spielt einen Ordner, der per Dialog ausgewäht wird, ab. Berücksichtigt auch Playlisten.
    ' Unterordner werden ebenfalls berücksichtigt.
    Public Function QuickplayDir(Optional ByVal DirPath As String = "", Optional ByVal DontPlay As Boolean = False, Optional ByVal SilentMode As Boolean = False) As Boolean
        If DirPath = "" Then
            Dim FolderDialog As New System.Windows.Forms.FolderBrowserDialog()
            FolderDialog.ShowNewFolderButton = False
            FolderDialog.Description = Loc("Bitte wählen Sie einen Ordner aus. Unterordner werden berücksichtigt; Playlisten nur nach Bestätigung.")

            FolderDialog.ShowDialog()
            If FolderDialog.SelectedPath = "" Then Return False

            DirPath = FolderDialog.SelectedPath
        End If

        Try
            Dim UsePlaylists As Boolean = False
            Dim AskedForPlaylists As Boolean = False
            Dim StartPos As Integer = Playlist.colMediaIDs.Count ' Speichert die Position des momentan letzten Playlist-Elements

            If SilentMode Then
                AskedForPlaylists = True
                UsePlaylists = True ' Playlists werden berücksichtigt
            End If

            If SilentMode = False Then ShowNotify(Loc("Analysiere Ordner..."))
            Debug.Print("start searching directory: " + Now.ToString + ":" + Now.Millisecond.ToString)
            For Each Filename In My.Computer.FileSystem.GetFiles(DirPath, FileIO.SearchOption.SearchAllSubDirectories)
                If IsPlaylist(Filename) Then
                    If AskedForPlaylists = False Then
                        UsePlaylists = CBool(MsgBox(Loc("Es wurde eine Playliste gefunden. Sollen Playlisten berücksichtigt werden?"), MsgBoxStyle.YesNo + MsgBoxStyle.Question, Loc("Frage")) = MsgBoxResult.Yes)
                        AskedForPlaylists = True
                    End If
                    ' Sollte die Datei eine Playlist sein, wird sie geladen - sofern der Benutzer dem zugestimmt hat - und die einzelnen Elemente dieser Playliste der virtuellen Playliste hinzugefügt.
                    If UsePlaylists Then LoadPlaylist(Filename)
                Else
                    If IsMediaFile(Filename) Then AddToPlaylist(LoadMedium(Filename)) ' Zur virtuellen Playliste hinzufügen - aber nur bekannte Formate!
                End If
            Next
            Debug.Print("end searching directory: " + Now.ToString + ":" + Now.Millisecond.ToString)

            If DontPlay = False And Playlist.colMediaIDs.Count > StartPos + 1 Then ' Diese Bedingung ist notwendig, da ja z.B. eine leere Playliste keine neue Elemente zur virtuellen Playliste hinzufügen würde.
                PlaylistGoTo(StartPos + 1, True)  ' +1 , da Index/Pos. des ehemals letzten Elements. Spielt das Medium ab.
            End If

            Return True
        Catch ex As Exception
            Log("QuickplayDir()", Loc("Fehler beim Abspielen eines Ordners.") + " (" + DirPath + ")", True)
            Return False
        End Try
    End Function

    ' Lädt eine "FileList" und spielt sie ab.
    ' FileList: Array aus Dateinamen (Mediendateien, Playlisten, Ordner (mit Unterordnern), URLs)
    ' Dies ist die mächtigste Qucikplay*-Methode.
    Public Sub QuickplayFilelist(ByVal FileList() As String)
        Dim Filename As String

        Dim StartPos As Integer = Playlist.colMediaIDs.Count  ' Speichert die Position des momentan letzten Playlist-Elements

        For Each Filename In FileList
            Debug.Print(Filename)
            Try
                If My.Computer.FileSystem.GetFileInfo(Filename).Exists = False Then ' Wenn false, handelt es sich vermutlich um einen Ordner.
                    If My.Computer.FileSystem.GetDirectoryInfo(Filename).Exists Then
                        Debug.Print("Ordner: " + Filename)

                        QuickplayDir(Filename, True, True)
                    Else
                        ' ... wird ignoriert.
                    End If
                Else ' Datei:
                    Debug.Print("Datei: " + Filename)

                    If IsPlaylist(Filename) Then
                        ' Sollte die Datei eine Playlist sein, wird sie geladen und die einzelnen Elemente dieser Playliste werden der virtuellen Playliste hinzugefügt.
                        LoadPlaylist(Filename)
                    Else
                        If IsMediaFile(Filename) Or IsURL(Filename) Then AddToPlaylist(LoadMedium(Filename)) ' Zur virtuellen Playliste hinzufügen.
                    End If
                End If
            Catch ex As Exception
                Debug.Print(ex.Message)
            End Try
        Next

        If Playlist.colMediaIDs.Count > StartPos Then ' Diese Bedingung ist notwendig, da ja z.B. eine leere Playliste keine neue Elemente zur virtuellen Playliste hinzufügen würde.
            PlaylistGoTo(StartPos + 1, True)  ' +1 , da Index/Pos. des ehemals letzten Elements. Spielt das Medium ab.
        End If
    End Sub

    ' Spielt ein existierendes Medium ab.
    ' Wenn das Medium bereits abgespielt wird, wird (bei Soundmedien) zum Anfang zurückgesprungen.
    ' Das eigentliche Abspielen wird an eine Submethode der passenden Engine delegiert.
    Public Function PlayMedium(ByVal MediumID As Integer) As Boolean
        Dim Medium As sMedium
        Dim IsOkay As Boolean

        If MediumID < 0 Or MediumID > Media.Count Or Media.Count = 0 Then
            Err.Raise(514, , "Invalid MediumID: " + MediumID.ToString)
            Exit Function
        End If
        Medium = Media(MediumID)

        Select Case Medium.MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                IsOkay = PlaySound(MediumID)
                Medium = Media(MediumID) ' NOTWENDIG! Da zwischendurch Eigenschaften des Mediums verändert werden, muss dieses Instanz hier geupdatet werden.
                If Medium.PlayLen = IsVoid And Not (Medium.MediumType = eMediumTypes.eNetstream Or Medium.MediumType = eMediumTypes.eError) Then
                    Medium.PlayLen = GetSoundPlayLen(MediumID)
                    Media(MediumID) = Medium
                End If
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
            Case eMediumTypes.eUndefined
                Log("PlayMedium", Loc("Versuche, Medium unbekannten Typs abzuspielen.") + "( " + Medium.Filename + " )")
                IsOkay = PlaySound(MediumID)
        End Select

        If IsOkay Then
            If Not Playlist.CurMediumID = MediumID Then
                Dim Index As Integer = GetPlaylistPosition(MediumID) ' ist das Medium schon in der virtuellen Playliste?
                If Index = IsVoid Then
                    PlaylistGoTo(AddToPlaylist(MediumID) + 1, , True) ' Medium der virtuellen Playliste hinzufügen
                Else
                    PlaylistGoTo(Index + 1, , True)
                End If
            End If
            Playlist.IsForwarding = True

            ShowNotify(Medium.GetBestTitle())
        Else
            ShowNotify(Loc("Kann Medium nicht abspielen."))
        End If

        UpdateGUI(True)
        Return IsOkay
    End Function

    ' Ruft die Funktion PlayMedium() auf; aber anhand der Position eines Mediums in der virtuellen Playliste.
    ' Gibt true zurück, wenn das Medium erfolgreich abgespielt werden konnte.
    Public Function PlayMedium_ByPlaylistPosition(ByVal PlaylistPosition As Integer)
        If PlaylistPosition < 1 Or PlaylistPosition > Playlist.colMediaIDs.Count Then
            Err.Raise(514, , "Invalid command / invalid value, Sub PlaylistGoTo()")
            Return False
        End If

        Dim Index = PlaylistPosition - 1 ' Umrechnung, der Index ist nullbasiert
        If Not Index = Playlist.CurMediumIndex Then
            PlaylistGoTo(PlaylistPosition, , True)
        End If
        Return PlayMedium(Playlist.CurMediumID)
    End Function

    ' Spielt einen Favorit ab (und lädt ihn vorher ggf.).
    ' Dazu wird lediglich die ID des Favorits benötigt.
    Public Sub PlayFavorit(ByVal FavoritID As Integer)
        Dim Target As String = Favorits(FavoritID).Target

        Dim StartPos As Integer = Playlist.colMediaIDs.Count ' Speichert die Position des momentan letzten Playlist-Elements

        If IsPlaylist(Target) Then
            ' Sollte die Datei eine Playliste sein, wird sie geladen und die einzelnen Elemente dieser Playliste werden der virtuellen Playliste hinzugefügt.
            LoadPlaylist(Target)
        Else
            AddToPlaylist(LoadMedium(Target)) ' Zur virtuellen Playliste hinzufügen.
        End If

        If Playlist.colMediaIDs.Count > StartPos Then ' Diese Bedingung ist notwendig, da ja z.B. eine leere Playliste keine neue Elemente zur virtuellen Playliste hinzufügen würde.
            PlaylistGoTo(StartPos + 1, True) ' +1 , da Index/Pos. des ehemals letzten Elements. Spielt das Medium ab.
        End If
    End Sub

    ' Beendet ggf. die Wiedergabe des aktuellen Mediums
    Public Sub StopPlaying()
        If Playlist.CurMediumID = IsVoid Then Exit Sub

        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                StopSound()
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
        End Select
        Playlist.IsForwarding = False
    End Sub

    ' Deaktiviert oder aktiviert die Tonausgabe die Wiedergabe
    ' (diese werden aber unabhängig davon weiter abgespielt)
    Public Sub SetMuteMode(ByVal Mute As Boolean)
        If Playlist.CurMediumID = IsVoid Then Exit Sub

        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                afi.SetSoundMuteMode(Mute)
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
        End Select
    End Sub

    ' Fragt den Status der Tonausgabe für die Wiedergabe ab.
    Public Function GetMuteMode() As Boolean
        If Playlist.CurMediumID = IsVoid Then Exit Function

        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                Return afi.GetSoundMuteMode()
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
        End Select
    End Function

    ' Legt die Lautstärke der Tonausgabe fest
    ' Volume: 0-100 %
    ' Evtl. wird der Lautstärkewert erst in UserVolume zwischengespeichert und erst später wirklich gesetzt.
    Public Function SetVolume(ByVal Volume As Single) As Boolean
        UserVolume = Volume

        If Playlist.CurMediumID = IsVoid Then Return False
        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                Return afi.SetSoundVolume(Volume)
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                Return True
            Case Else
                Return True
        End Select
    End Function

    ' Fragt die (User-)Lautstärke der Tonausgabe von Audio-Medien ab
    ' Volume: 0-100 %
    Public Function GetVolume() As Single
        Return UserVolume
    End Function

    ' Erhöht die Lautstärke um einen Schritt.
    ' Die Schrittgröße ist als Konstante hinterlegt.
    ' Die Lautstärke wird nicht über 100% erhöht.
    Public Function IncreaseVolume() As Boolean
        Dim Vol As Single
        Vol = GetVolume() + VolumeStep
        If Vol > 100 Then Vol = 100
        Return SetVolume(Vol)
    End Function

    ' Verringert die Lautstärke um einen Schritt.
    ' Die Schrittgröße ist als Konstante hinterlegt.
    ' Die Lautstärke wird nicht unter 0 % verringert.
    Public Function DecreaseVolume() As Boolean
        Dim Vol As Single
        Vol = GetVolume() - VolumeStep
        If Vol < 0 Then Vol = 0
        Return SetVolume(Vol)
    End Function

    ' Fügt ein Medium der virtuellen Playlist hinzu, auch wenn es bereits in ihr enthalten ist.
    ' Gibt den Index (~ die Position) des Mediums in der Playliste zurück.
    ' Achtung: Die GUI wird nur geupdatet, wenn das extra via Parameter angegeben wird!
    Public Function AddToPlaylist(ByVal MediumID As Integer, Optional ByVal UpdateGUI As Boolean = False) As Integer
        If MediumID < 0 Or MediumID > Media.Count Or Media.Count = 0 Then
            Err.Raise(514, , "Invalid MediumID: " + MediumID.ToString)
            Return IsVoid
        End If

        Dim Medium As sMedium = Media(MediumID)
        Media(MediumID) = Medium

        Playlist.colMediaIDs.Add(CInt(MediumID))
        If UpdateGUI Then askin.UpdateGUI(False, False, True) ' GUI updaten

        Return Playlist.colMediaIDs.Count - 1
    End Function

    ' Gibt den Index (~ die Position) eines Mediums in der virtuellen Playliste zurück.
    ' ACHTUNG: Es wird der nullbasierte INDEX zurückgegeben.
    ' Berechnung der logischen Position: Position = Index + 1
    ' Gibt die Konstante IsVoid zurück, wenn das Medium nicht in der Playliste ist.
    Public Function GetPlaylistPosition(ByVal MediumID) As Integer
        If MediumID < 0 Or MediumID > Media.Count Or Media.Count = 0 Then
            Err.Raise(514, , "Invalid MediumID: " + MediumID.ToString)
            Return IsVoid
        End If

        Dim i As Integer
        For i = 0 To Playlist.colMediaIDs.Count - 1
            If Playlist.colMediaIDs(i) = MediumID Then Return i
        Next
        Return IsVoid ' MediumID nicht gefunden
    End Function

    ' Leert die virtuelle Playliste
    Public Sub ClearPlaylist()
        StopPlaying() ' Wiedergabe beenden

        Playlist.colMediaIDs.Clear()
        Playlist.CurMediumID = IsVoid
        Playlist.CurMediumIndex = IsVoid

        UpdateGUI(True)
    End Sub

    ' Speichert eine Playliste.
    ' Wird kein Dateiname übergeben, wird ein Datei-Speichern-Dialog angezeigt.
    Public Function SavePlaylist(Optional ByVal Filename As String = "") As Boolean
        If Filename = "" Then
            Dim SaveDialog As New SaveFileDialog
            SaveDialog.Filter = PlaylistDialogFilter
            SaveDialog.ShowDialog()

            If SaveDialog.FileName = "" Then
                Return False
            Else
                Filename = SaveDialog.FileName
            End If
        End If

        Select Case GetFilenameExtension(Filename).ToUpper
            Case "M3U"
                Return Playlist_M3U_Export(Filename)
            Case "PLS"
                Return Playlist_PLS_Export(Filename)
            Case "WPL"
                Return Playlist_WPL_Export(Filename)
            Case "XSPF"
                Return Playlist_XSPF_Export(Filename)
            Case "TXT"
                Return Playlist_TXT_Export(Filename)
            Case Else

        End Select
    End Function

    ' Lädt eine Playlist, spielt sie jedoch nicht ab.
    ' Wird kein Dateiname übergeben, so wird ein Datei-Öffnen-Dialog angezeigt.
    ' Das eigentliche Laden geschieht in einer anderen Methode.
    Public Function LoadPlaylist(Optional ByVal Filename As String = "", Optional ByVal ClearCurrentPlaylist As Boolean = False) As Boolean
        If Filename = "" Then
            Dim OpenDialog As New OpenFileDialog

            OpenDialog.Filter = PlaylistDialogFilter
            OpenDialog.CheckFileExists = True
            OpenDialog.AddExtension = True
            OpenDialog.ShowDialog()

            If OpenDialog.FileName = "" Then
                Return False
            Else
                Filename = OpenDialog.FileName
            End If
        End If

        Dim IsOkay As Boolean = False

        If ClearCurrentPlaylist Then ClearPlaylist()

        Select Case GetFilenameExtension(Filename).ToUpper
            Case "M3U"
                IsOkay = Playlist_M3U_Import(Filename)
            Case "PLS"
                IsOkay = Playlist_PLS_Import(Filename)
            Case "WPL"
                IsOkay = Playlist_WPL_Import(Filename)
            Case "XSPF"
                IsOkay = Playlist_XSPF_Import(Filename)
            Case "TXT"
                IsOkay = Playlist_TXT_Import(Filename)
            Case Else

        End Select
        If IsOkay Then
            If Playlist.CurMediumID = IsVoid Or ClearCurrentPlaylist Then ' Wenn Playliste leer war oder durch eine neue ersetzt wurde:
                If Playlist.colMediaIDs.Count > 0 Then
                    PlaylistGoTo(1, False, True)  ' 1. Medium in der Playlist zum aktuellen machen
                Else
                    Playlist.CurMediumIndex = IsVoid ' Playlist ist leer, kann kein Medium zum aktuellen machen
                    Playlist.CurMediumID = IsVoid
                End If
            End If

            If ClearCurrentPlaylist Then
                UpdateGUI(True)
            Else
                UpdateGUI(False, False, True)
            End If
        End If

        Return IsOkay
    End Function

    ' Importiert eine Playliste im "Extendend M3U Format" oder im gewöhnlichen (veralteten) M3U-Format.
    Public Function Playlist_M3U_Import(ByVal Filename As String) As Boolean
        Dim sr As IO.StreamReader
        Dim Line As String, Title As String = "", Path As String = GetPath(Filename)
        Dim Pos As Integer

        Try
            sr = New IO.StreamReader(Filename)

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                If strContains(Line, "/") = True And strContains(Line, "//") = False Then ' Linux-M3U-Playliste ("//" tritt bei Net-Streams auf)
                    sr.Close()
                    Return False ' Abbrechen.
                End If
                If Line.StartsWith("#") Then ' Hinweis: Mit # werden Zeilen mit Extra-Informationen zu den Dateien eingeleitet. Diese werden hier unter bestimmten Umständen ausgewertet (Audia kann diese Informationen aber auch selbst erzeugen)
                    Pos = InStr(Line, ",") ' es kann auch ausdrücklich mit "#" beginnende Zeilen OHNE "," geben!
                    If Pos > 0 Then Title = Mid(Line, Pos + 1) ' Hier steht der (Song-)Titel
                End If
                If Not (Line.StartsWith("#") Or Trim(Line) = "") Then
                    Line = GetAbsolutFilename(Path, Line) ' Beispiel: Die Playliste befindet sich in "C:\Musik\Caendoo\" und Zeile="Yeanui.mp3", dann wird Zeile zu "C:\Musik\Caendoo\Yeanui.mp3".
                    If strContains(Line, "/") = False Then  ' Das Slash kennzeichnet Net-Streams, deren Existenz nicht so einfach geprüft werden kann. Hinweis: Der Fall, das "/" gar keinen Stream kennzeichnet, wurde oben schon ausgeschlossen.
                        If IO.File.Exists(Line) Then
                            Pos = AddToPlaylist(LoadMedium(Line)) ' Zur virtuellen Playliste hinzufügen
                        End If
                    Else
                        Pos = AddToPlaylist(LoadMedium(Line)) ' Zur virtuellen Playliste hinzufügen
                    End If
                    If Title <> "" Then
                        If Media(Playlist.colMediaIDs(Pos)).Title = "" Then ' WICHTIG: Den Titel nur übernehmen, wenn das Medium noch keinen besitzt. Denn in einer Playliste könnte etwas falsches stehen!
                            Dim Medium As sMedium = Media(Playlist.colMediaIDs(Pos))
                            Medium.Title = Title
                            Media(Playlist.colMediaIDs(Pos)) = Medium
                        End If
                        Title = ""
                    End If
                End If
            Loop

            sr.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_M3U_Import", Loc("Fehler beim Laden einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Exportiert die virtuelle Playliste in das "Extendend M3U Format".
    Public Function Playlist_M3U_Export(ByVal Filename As String) As Boolean
        Dim sw As IO.StreamWriter
        Dim i As Integer

        Try
            sw = New IO.StreamWriter(Filename)

            sw.WriteLine("#EXTM3U") ' leitet eine M3U-Playliste ein
            For i = 0 To Playlist.colMediaIDs.Count - 1
                sw.WriteLine("#EXTINF:0," + Media(Playlist.colMediaIDs(i)).GetBestTitle) ' (Song-)Titel speichern
                '                     0=Zeitlänge des Mediums in Sekunden (0 schreibt auch der Windows Mediaplayer 9!)
                sw.WriteLine(Media(Playlist.colMediaIDs(i)).Filename) ' Dateiname speichern
                sw.WriteLine("") ' Leerzeile wegen Kompatiblität zum Nero Mediaplyer erforderlich (andere Player benötigen diese Leerzeile nicht)
            Next

            sw.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_M3U_Export", Loc("Fehler beim Speichern einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Importiert eine Playliste im "PLS Format, Version 2".
    Public Function Playlist_PLS_Import(ByVal Filename As String) As Boolean
        Dim sr As IO.StreamReader
        Dim Line As String, Title As String = "", Path As String = GetPath(Filename)
        Dim Length As Integer

        Try
            sr = New IO.StreamReader(Filename)

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                If Line.StartsWith("File") Then
                    Line = Mid(Line, InStr(Line, "=") + 1)
                    Line = GetAbsolutFilename(Path, Line) ' Beispiel: Die Playliste befindet sich in "C:\Musik\Caendoo\" und Zeile="Yeanui.mp3", dann wird Zeile zu "C:\Musik\Caendoo\Yeanui.mp3".
                    If strContains(Line, "/") = False Then  ' Das Slash kennzeichnet Net-Streams, deren Existenz nicht so einfach geprüft werden kann.
                        If IO.File.Exists(Line) Then
                            AddToPlaylist(LoadMedium(Line)) ' Zur virtuellen Playliste hinzufügen
                        End If
                    Else
                        AddToPlaylist(LoadMedium(Line)) ' Zur virtuellen Playliste hinzufügen
                    End If
                End If
                If Line.StartsWith("Length") Then
                    Length = Val(Mid(Line, InStr(Line, "=") + 1)) ' Länge (in Sekunden) extrahieren
                    If Length > 0 Then ' Nur wenn Länge bekannt:
                        Dim Medium As sMedium = Media(Playlist.colMediaIDs(Playlist.colMediaIDs.Count - 1))
                        If Medium.PlayLen = IsVoid Then ' Nur, wenn Längenangabe des Mediums unbekannt (denn die Länge in der Playliste könnte ja falsch sein):
                            Medium.PlayLen = Length
                            Media(Playlist.colMediaIDs(Playlist.colMediaIDs.Count - 1)) = Medium
                        End If
                    End If
                End If
                If Line.StartsWith("Title") Then
                    Title = Mid(Line, InStr(Line, "=") + 1) ' Titel extrahieren
                    If Not Title = "" Then ' Nur wenn Titel wirklich angegeben:
                        Dim Medium As sMedium = Media(Playlist.colMediaIDs(Playlist.colMediaIDs.Count - 1))
                        If Medium.Title = "" Then ' Nur, wenn Titel des Mediums unbekannt (denn der Titel in der Playliste könnte ja falsch sein):
                            Medium.Title = Title
                            Media(Playlist.colMediaIDs(Playlist.colMediaIDs.Count - 1)) = Medium
                        End If
                    End If
                End If
            Loop

            sr.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_PLS_Import", Loc("Fehler beim Laden einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Exportiert die virtuelle Playliste in das "PLS Format, Version 2".
    Public Function Playlist_PLS_Export(ByVal Filename As String) As Boolean
        Dim sw As IO.StreamWriter
        Dim i As Integer

        Try
            sw = New IO.StreamWriter(Filename)

            sw.WriteLine("[playlist]") ' leitet eine PLS-Playliste ein
            sw.WriteLine("NumberOfEntries=" + Playlist.colMediaIDs.Count.ToString) ' Anzahl der Einträge
            sw.WriteLine("") ' Leerzeile aus Kompatiblitätgründen einfügen (einige Player benötigen diese Leerzeile nicht)
            For i = 0 To Playlist.colMediaIDs.Count - 1
                sw.WriteLine("File" + (i + 1).ToString + "=" + Media(Playlist.colMediaIDs(i)).Filename) ' Dateiname speichern
                sw.WriteLine("Title" + (i + 1).ToString + "=" + Media(Playlist.colMediaIDs(i)).GetBestTitle()) ' (Song-)Titel speichern
                sw.WriteLine("Length" + (i + 1).ToString + "=" + CInt(Media(Playlist.colMediaIDs(i)).PlayLen).ToString) ' Abspiellänge speichern, in "ganzen" Sekunden
                sw.WriteLine("") ' Leerzeile aus Kompatiblitätgründen einfügen (einige Player benötigen diese Leerzeile nicht)
            Next
            sw.Write("Version=2") ' Version des PLS-Formats

            sw.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_PLS_Export", Loc("Fehler beim Speichern einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Importiert eine Playliste im "Windows Mediaplayer Playlist Format".
    Public Function Playlist_WPL_Import(ByVal Filename As String) As Boolean
        Try
            Dim reader As Xml.XmlReader
            Dim Element As String, File As String, Path As String = GetPath(Filename)

            reader = Xml.XmlReader.Create(Filename)

            While reader.Read()
                If reader.NodeType = Xml.XmlNodeType.Element Then
                    Element = reader.Name.ToLower
                    If reader.HasAttributes And Element = "media" Then
                        reader.MoveToFirstAttribute()
                        Do
                            If reader.Name.ToLower = "src" Then
                                File = GetAbsolutFilename(Path, reader.Value) ' Beispiel: Die Playliste befindet sich in "C:\Musik\Caendoo\" und Zeile="Yeanui.mp3", dann wird Zeile zu "C:\Musik\Caendoo\Yeanui.mp3".
                                If strContains(File, "/") = False Then  ' Das Slash kennzeichnet Net-Streams, deren Existenz nicht so einfach geprüft werden kann.
                                    If IO.File.Exists(File) Then
                                        AddToPlaylist(LoadMedium(File)) ' Zur virtuellen Playliste hinzufügen
                                    End If
                                Else
                                    AddToPlaylist(LoadMedium(File)) ' Zur virtuellen Playliste hinzufügen
                                End If
                            End If
                        Loop Until reader.MoveToNextAttribute() = False
                    End If
                End If
            End While

            reader.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_WPL_Import", Loc("Fehler beim Laden einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Exportiert eine Playliste im "Windows Mediaplayer Playlist Format".
    ' Achtung: Das Format wird möglicherweise nicht zu 100% eingehalten.
    ' Der Windows Media Player 12 kann es aber fehlerfrei lesen.
    Public Function Playlist_WPL_Export(ByVal Filename As String) As Boolean
        Try
            Dim writer As Xml.XmlWriter
            Dim settings As Xml.XmlWriterSettings

            settings = New Xml.XmlWriterSettings()
            settings.Indent = True
            writer = Xml.XmlWriter.Create(Filename, settings)

            writer.WriteStartDocument()
            writer.WriteStartElement("smil")
            ' Header:
            writer.WriteStartElement("head")
            writer.WriteStartElement("meta")
            writer.WriteAttributeString("name", "Generator")
            writer.WriteAttributeString("content", Application.ProductName)
            writer.WriteEndElement()
            writer.WriteStartElement("meta")
            writer.WriteAttributeString("name", "ItemCount")
            writer.WriteAttributeString("content", Playlist.colMediaIDs.Count.ToString)
            writer.WriteEndElement()
            writer.WriteStartElement("title")
            writer.WriteValue(GetFileTitle(Filename)) ' Titel der Playliste; hier wird einfach der Dateititel verwendet
            writer.WriteEndElement()
            writer.WriteEndElement()
            ' Body:
            writer.WriteStartElement("body")
            writer.WriteStartElement("seq")
            For Each MediumID In Playlist.colMediaIDs
                writer.WriteStartElement("media")
                writer.WriteAttributeString("src", Media(MediumID).Filename)
                writer.WriteEndElement()
            Next
            writer.WriteEndElement()
            writer.WriteEndElement()

            writer.WriteEndElement()
            writer.WriteEndDocument()
            writer.Close()

            Return True
        Catch ex As Exception
            Log("Playlist_WPL_Export", Loc("Fehler beim Speichern einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Importiert eine Playliste im "XML Shareable Playlist Format".
    Public Function Playlist_XSPF_Import(ByVal Filename As String) As Boolean
        Try
            Dim reader As Xml.XmlReader
            Dim Element As String = "", File As String, Title As String = "", Path As String = GetPath(Filename)
            Dim Pos As Integer

            reader = Xml.XmlReader.Create(Filename)

            While reader.Read()
                If reader.NodeType = Xml.XmlNodeType.Element Then
                    Element = reader.Name.ToLower
                End If
                If reader.NodeType = Xml.XmlNodeType.Text And Element = "location" Then
                    File = reader.Value
                    If File.ToLower.StartsWith("file:///") Then
                        File = Mid(File, 9)
                        File = Replace(File, "/", "\") ' Notwendig. Das Format verlangt diese Ersetzung bei Dateinamen.
                        File = Replace(File, "%20", " ") ' Notwendig. Das Format verlangt diese Ersetzung bei Dateinamen.
                    End If
                    File = GetAbsolutFilename(Path, File) ' Beispiel: Die Playliste befindet sich in "C:\Musik\Caendoo\" und Zeile="Yeanui.mp3", dann wird Zeile zu "C:\Musik\Caendoo\Yeanui.mp3".
                    If strContains(File, "/") = False Then  ' Das Slash kennzeichnet Net-Streams, deren Existenz nicht so einfach geprüft werden kann.
                        If IO.File.Exists(File) Then
                            Pos = AddToPlaylist(LoadMedium(File)) ' Zur virtuellen Playliste hinzufügen
                        End If
                    Else
                        Pos = AddToPlaylist(LoadMedium(File)) ' Zur virtuellen Playliste hinzufügen
                    End If
                    If Title <> "" Then
                        If Media(Playlist.colMediaIDs(Pos)).Title = "" Then ' WICHTIG: Den Titel nur übernehmen, wenn das Medium noch keinen besitzt. Denn in einer Playliste könnte etwas falsches stehen!
                            Dim Medium As sMedium = Media(Playlist.colMediaIDs(Pos))
                            Medium.Title = Title
                            Media(Playlist.colMediaIDs(Pos)) = Medium
                        End If
                        Title = ""
                    End If
                End If
                If reader.NodeType = Xml.XmlNodeType.Text And Element = "title" Then
                    Title = reader.Value
                End If
            End While

            reader.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_XFPS_Import", Loc("Fehler beim Laden einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Exportiert eine Playliste im "XML Shareable Playlist Format".
    ' Dies kann z.B. vom VLC gelesen werden.
    ' Achtung: Das Format wird möglicherweise nicht zu 100% eingehalten.
    Public Function Playlist_XSPF_Export(ByVal Filename As String) As Boolean
        Try
            Dim writer As Xml.XmlWriter
            Dim settings As Xml.XmlWriterSettings
            Dim value As String

            settings = New Xml.XmlWriterSettings()
            settings.Indent = True
            writer = Xml.XmlWriter.Create(Filename, settings)

            writer.WriteStartDocument()
            writer.WriteStartElement("playlist")
            writer.WriteAttributeString("version", "1")
            writer.WriteAttributeString("xlmns", "http://xspf.org/ns/0/")
            writer.WriteStartElement("tracklist")
            ' Medien:
            For Each MediumID In Playlist.colMediaIDs
                writer.WriteStartElement("track")

                writer.WriteStartElement("title") ' Title-Informationen sind vom Format her optional (müssen also nicht ubnedingt geschrieben werden)
                writer.WriteValue(Media(MediumID).GetBestTitle)
                writer.WriteEndElement()

                writer.WriteStartElement("location")
                value = Media(MediumID).Filename
                If IsURL(Media(MediumID).Filename) = False Then ' Bei "echten" Dateinamen:
                    value = Replace(value, "\", "/") ' Notwendig. Das Format verlangt diese Ersetzung bei Dateinamen.
                    value = Replace(value, " ", "%20") ' Notwendig. Das Format verlangt diese Ersetzung bei Dateinamen.
                    value = "file:///" + value ' Dieses Präfix vor "echten" Dateinamen wird vom Format gefordert.
                End If
                writer.WriteValue(value)
                writer.WriteEndElement()

                writer.WriteEndElement()
            Next
            writer.WriteEndElement()
            writer.WriteEndElement()
            writer.WriteEndDocument()
            writer.Close()

            Return True
        Catch ex As Exception
            Log("Playlist_XFPS_Export", Loc("Fehler beim Speichern einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Importiert eine Playliste im "TXT-Format".
    ' Darunter wird einfach nur eine Auflistung von Dateinamen/URLS verstanden (ein Eintrag pro Zeile).
    Public Function Playlist_TXT_Import(ByVal Filename As String) As Boolean
        Dim sr As IO.StreamReader
        Dim Line As String, Path As String = GetPath(Filename)

        Try
            sr = New IO.StreamReader(Filename)

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                Line = GetAbsolutFilename(Path, Line) ' Beispiel: Die Playliste befindet sich in "C:\Musik\Caendoo\" und Zeile="Yeanui.mp3", dann wird Zeile zu "C:\Musik\Caendoo\Yeanui.mp3".
                If strContains(Line, "/") = False Then  ' Das Slash kennzeichnet Net-Streams, deren Existenz nicht so einfach geprüft werden kann.
                    If IO.File.Exists(Line) Then
                        AddToPlaylist(LoadMedium(Line)) ' Zur virtuellen Playliste hinzufügen
                    End If
                Else
                    AddToPlaylist(LoadMedium(Line)) ' Zur virtuellen Playliste hinzufügen
                End If
            Loop

            sr.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_TXT_Import", Loc("Fehler beim Laden einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Exportiert die virtuelle Playliste in das "TXT-Format".
    Public Function Playlist_TXT_Export(ByVal Filename As String) As Boolean
        Dim sw As IO.StreamWriter
        Dim i As Integer

        Try
            sw = New IO.StreamWriter(Filename)

            For i = 0 To Playlist.colMediaIDs.Count - 1
                sw.WriteLine(Media(Playlist.colMediaIDs(i)).Filename) ' Dateiname speichern
            Next

            sw.Close()
            Return True
        Catch ex As Exception
            Log("Playlist_TXT_Export", Loc("Fehler beim Speichern einer Playliste: ") + Filename, True)
            Return False
        End Try
    End Function

    ' Dies ist der Haupttimer. Er kümmert sich um alle wichtigen, periodischen Vorgänge wie z.B. die Aktualisierung der GUI.
    Private Sub MainTimer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles MainTimer.Tick
        Dim ForceUpdateAll As Boolean = False  ' Komplettes GUI-Update erzwingen?

        If LockMainTimer Then ' Wenn true, wird diese Prozedur noch ausgeführt. Doppelte Ausführung soll so verhindert werden.
            Log("Maintenance", Loc("Warnung: Zugriffversuch auf gesperrte Hauptkontrollprozedur."))
            Exit Sub
        Else
            LockMainTimer = True
        End If

        If Playlist.IsForwarding Then
            If Media(Playlist.CurMediumID).MediumType = eMediumTypes.eNetstream Then ' Netstreams werden idR unendlich abgespielt
                Dim OpenState As FMOD.OPENSTATE
                Dim Buffered As Integer
                Dim Starving As Boolean
                Media(Playlist.CurMediumID).FModSound.getOpenState(OpenState, Buffered, Starving) ' Status des Streams ermitteln
                If OpenState = FMOD.OPENSTATE.ERROR Then ' Verbindung wegen Fehler abgebrochen/nicht möglich:
                    Log("Maintenance", Loc("Fehlerhafte Verbindung zum Stream"))
                    ShowNotify(Loc("Fehlerhafte Verbindung zum Stream."))
                    GoTo TryForwarding ' Ja, ein böses GoTo! :D
                End If
                If Starving Then ' Verbindungsprobleme
                    ShowNotify(Loc("Verbindung zum Stream verloren."))
                    StopPlaying()
                Else
                    If Media(Playlist.CurMediumID).IsPlayLenTested And Media(Playlist.CurMediumID).HasPlayLen Then
                        GoTo TestForwarding
                    End If
                End If
            Else
TestForwarding:
                If GetPlayPos() = GetMediumPlayLen(Playlist.CurMediumID) Or (IntroMode And (GetPlayPos() > IntroModeDuration)) Then ' Lied ist am Ende angekommen
TryForwarding:
                    If RepeatMedium Then
                        PlayMedium(Playlist.CurMediumID) ' akt. Medium erneut abspielen
                    Else
                        If RandomMode Then
                            If Playlist.colMediaIDs.Count = 1 Then
                                PlayMedium_ByPlaylistPosition(1)
                            Else
                                Dim Index As Integer
                                Do
                                    Index = Rnd.Next(0, Playlist.colMediaIDs.Count - 1)
                                Loop Until Index <> Playlist.CurMediumIndex ' Es soll nicht das gleiche Medium zwei Mal hintereinander abgespielt werden.
                                PlayMedium_ByPlaylistPosition(Index + 1)
                            End If
                        Else
                            If Playlist.CurMediumIndex < Playlist.colMediaIDs.Count - 1 Or RepeatPlaylist Then
                                Playlist.CurMediumIndex += 1
                                If Playlist.CurMediumIndex = Playlist.colMediaIDs.Count Then Playlist.CurMediumIndex = 0 ' Ende der Playliste erreicht, von vorne beginnen!
                                Playlist.CurMediumID = Playlist.colMediaIDs(Playlist.CurMediumIndex) ' MediumID ermitteln
                                PlayMedium(Playlist.CurMediumID)
                                ForceUpdateAll = True ' Komplettes GUI-Update erzwingen
                            Else
                                Playlist.IsForwarding = False ' Ende der Playliste UND der Wiedergabe erreicht, keine automatische neue Wiedergabe
                            End If
                        End If
                    End If
                End If
            End If
        End If

        CheckGlobalKeys()

        GlobalTicks += 1
        LockMainTimer = False

        UpdateGUI(ForceUpdateAll, True) ' GUI updaten
    End Sub

    ' Sucht nach neu angeschlossenen bzw. nicht mehr angeschlossenen portablen Datenträgern.
    Public Sub PortablesDetection()
        Dim di As IO.DriveInfo
        Dim Portable As sPortable
        Dim IsKnown As Boolean = False
        Dim Changed As Boolean = False
        Static IsLocked As Boolean = False
        Static NotMyFirstRun As Boolean = False ' Wird die Prozedur nicht zum 1.  Mal aufgerufen?

        If IsLocked Then Exit Sub

CheckOldPortables:
        ' Prüfe, ob vorhandene portable Datenträger noch aktiv/angeschlossen:
        For Each Portable In Portables
            IsKnown = False
            For Each di In IO.DriveInfo.GetDrives()
                If di.Name.ToUpper = Portable.Drive Then
                    IsKnown = True ' Datenträger ist noch angeschlossen
                    Exit For
                End If
            Next
            If IsKnown = False Then ' Der Datenträger ist nicht mehr verfügbar:
                ShowNotify(Portable.Name + " " + Loc("entfernt."))
                Portables.Remove(Portable) ' aus Liste entfernen
                Changed = True
                GoTo CheckOldPortables ' es kann nur ein Element aus der Liste entfernt werden, da es sonst zu einem Fehler kommen kann. Daher wird der Kontrollevorgang einfach wiederholt.
            End If
        Next

        ' Suche nach neuen portablen Datenträgern:
        For Each di In IO.DriveInfo.GetDrives()
            If di.DriveType = IO.DriveType.Removable Then
                For Each Portable In Portables
                    If Portable.Drive.ToUpper = di.Name.ToUpper Then
                        IsKnown = True ' Datenträger wurde bereits erkannt
                        Exit For
                    End If
                Next
                If IsKnown = False Then ' Es wurde ein portabler Datenträger angeschlossen/aktiviert.
                    IsLocked = True ' Erneuten Aufruf dieser Methode verhindert. Notwendig, weil evtl. erst auf eine Benutzereingabe gewartet werden muss.
                    Portable = New sPortable
                    Portable.Drive = di.Name.ToUpper
                    If My.Computer.FileSystem.FileExists(di.Name + PortableSettingsFilename) = False Then ' Ist der Datenträger noch nicht bekannt/katalogisiert?
                        Portable.Name = di.VolumeLabel
                        If MsgBox(Loc("Soll der folgende portable Datenträger von Audia verwaltet werden?") + vbCr + vbCr + di.Name + vbCr + "(" + di.VolumeLabel + ")", MsgBoxStyle.Question + MsgBoxStyle.YesNo, Loc("Portabler Datenträger entdeckt")) = MsgBoxResult.Yes Then
                            Portable.Enabled = True
                            Changed = True
                        Else
                            Portable.Enabled = False
                        End If
                        SetPortableSettings(di.Name + PortableSettingsFilename, Portable.Name, Portable.Enabled) ' Einstellungen auf Datenträger speichern
                    Else
                        GetPortableSettings(di.Name + PortableSettingsFilename, Portable.Name, Portable.Enabled) ' Einstellungen des Datenträgers auslesen
                        Changed = True
                        If NotMyFirstRun Then ShowNotify(Portable.Name + " " + Loc("angeschlossen.")) ' Beim Start des Players soll keine Meldung kommen, dass portable Datenträger gefunden wurde.
                    End If
                    Portable.Use = Portable.Enabled
                    Portables.Add(Portable) ' der Liste hinzufügen
                    IsLocked = False ' Erneuten Aufruf dieser Methode wieder erlauben.
                End If
            End If
        Next

        If Changed Then
            Dim i As Integer
            Dim ShowThisMenu As Boolean = False

            For i = frmMenu.menMenuPortables.DropDownItems.Count - 1 To 0 Step -1
                If Not frmMenu.menMenuPortables.DropDownItems(i).Tag = "" Then
                    frmMenu.menMenuPortables.DropDownItems.Remove(frmMenu.menMenuPortables.DropDownItems(i))
                End If
            Next

            For i = 0 To Portables.Count - 1
                Dim MenuItem As New ToolStripMenuItem
                MenuItem.Text = Portables(i).Name
                MenuItem.Tag = i.ToString
                MenuItem.Checked = Portables(0).Use
                frmMenu.menMenuPortables.DropDownItems.Insert(0, MenuItem)
                AddHandler frmMenu.menMenuPortables.DropDownItems(frmMenu.menMenuPortables.DropDownItems.Count - 1).Click, AddressOf askin.menUsePortable_Click ' Verknüpfung zur Ereignisbehandlung herstellen
                ShowThisMenu = True
            Next

            frmMenu.menMenuPortables.Enabled = ShowThisMenu
        End If

        NotMyFirstRun = True
    End Sub

    ' Einstellungen eines portablen Datenträgers auslesen.
    ' ACHTUNG: Die Ergebnisse werden in PortableName und PortableEnabled zurück gegeben!
    Public Function GetPortableSettings(ByVal PortableSettingsFile As String, ByRef PortableName As String, ByRef PortableEnabled As Boolean) As Boolean
        Dim sr As IO.StreamReader
        Dim Line As String

        Try
            sr = New IO.StreamReader(PortableSettingsFile)

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                If Line.ToLower.StartsWith("portablename=") Then PortableName = Mid(Line, 14)
                If Line.ToLower.StartsWith("portableenabled=") Then PortableEnabled = CBool(Mid(Line, 17))
            Loop

            sr.Close()
            Return True
        Catch ex As Exception
            Log("GetPortableSettings()", Loc("Fehler beim Auslesen der Einstellungen eines portablen Datenträgers: ") + PortableSettingsFile, True)
            Return False
        End Try
    End Function

    ' Einstellungen eines portablen Datenträgers speichern.
    Public Function SetPortableSettings(ByVal PortableSettingsFile As String, ByRef PortableName As String, ByRef PortableEnabled As Boolean) As Boolean
        Dim sw As IO.StreamWriter

        Try
            sw = New IO.StreamWriter(PortableSettingsFile)

            sw.WriteLine("' Audia Portable Settings File")
            sw.WriteLine("' Please do not remove nor edit this file!")
            sw.WriteLine("PortableName=" + PortableName)
            sw.WriteLine("PortableEnabled=" + PortableEnabled.ToString)

            sw.Close()
            Return True
        Catch ex As Exception
            Log("SetPortableSettings()", Loc("Fehler beim Speichern der Einstellungen eines portablen Datenträgers: ") + PortableSettingsFile, True)
            Return False
        End Try
    End Function

    ' Springt in der virtuellen Playliste ein Medium vorwärts.
    ' Kehrt ggf. an den Anfang der Playliste zurück.
    Public Sub PlaylistForward()
        If Playlist.CurMediumID = IsVoid Then Exit Sub

        Dim Pos As Integer = Playlist.CurMediumIndex
        Pos += 1 ' 1 Medium vorwärts
        If Pos > Playlist.colMediaIDs.Count - 1 Then Pos = 0

        PlaylistGoTo(Pos + 1)
    End Sub

    ' Springt in der virtuellen Playliste ein Medium zurück.
    ' Kehrt ggf. an das Ende der Playliste zurück.
    Public Sub PlaylistBack()
        If Playlist.CurMediumID = IsVoid Then Exit Sub

        Dim Pos As Integer = Playlist.CurMediumIndex
        Pos -= 1 ' 1 Medium vorwärts
        If Pos < 0 Then Pos = Playlist.colMediaIDs.Count - 1

        PlaylistGoTo(Pos + 1)
    End Sub

    ' Springt zu einer angegebenen Position in der virtuellen Playliste und spielt das Medium ab.
    ' PlaylistPosition: Zahl zwischen 1 und der Anzahl der Playlist-Elemente
    Public Sub PlaylistGoTo(ByVal PlaylistPosition As Integer, Optional ByVal ForcePlay As Boolean = False, Optional ByVal ForceDontPlay As Boolean = False)
        If PlaylistPosition < 1 Or PlaylistPosition > Playlist.colMediaIDs.Count Then
            Err.Raise(514, , "Invalid command / invalid value, Sub PlaylistGoTo()")
            Exit Sub
        End If

        If ForceDontPlay = True Or (ForcePlay = False And IsPlaying() = False) Then
            Playlist.CurMediumIndex = PlaylistPosition - 1
            Playlist.CurMediumID = Playlist.colMediaIDs(Playlist.CurMediumIndex)
        Else
            PlayMedium_ByPlaylistPosition(PlaylistPosition)
        End If
    End Sub

    ' Entfernt ein Medium, dessen Position in der Playlist bekannt ist, aus eben dieser.
    ' Der Parameter DontUpdate verhindert, dass die GUI anschließend geupdatet wird. Dies ist z.B. dann sinnvoll, wenn mehrere Elemente direkt hintereinander aus der Playliste entfernt werden. Die GUI muss dann anschließend "manuell" geupdatet werden.
    Public Sub RemoveFromPlaylist_ByPosition(ByVal PlaylistPosition As Integer, Optional ByVal DontUpdate As Boolean = False)
        If Playlist.CurMediumID = IsVoid Or PlaylistPosition < 1 Or PlaylistPosition > Playlist.colMediaIDs.Count Then
            'Stop
            Err.Raise(514, , "Invalid command / invalid value, Sub RemoveFromPlaylist")
            Exit Sub
        End If

        Dim IsCur As Boolean = False, WasPlaying As Boolean
        If Playlist.CurMediumIndex = PlaylistPosition - 1 Then IsCur = True
        WasPlaying = IsPlaying()

        If WasPlaying And IsCur Then StopPlaying()
        Playlist.colMediaIDs.RemoveAt(PlaylistPosition - 1)

        If Playlist.CurMediumIndex > Playlist.colMediaIDs.Count - 1 Then Playlist.CurMediumIndex -= 1
        If Playlist.colMediaIDs.Count = 0 Then
            Playlist.CurMediumIndex = IsVoid
            Playlist.CurMediumID = IsVoid
        Else
            Playlist.CurMediumID = Playlist.colMediaIDs(Playlist.CurMediumIndex)
        End If
        If WasPlaying Then
            If Playlist.CurMediumIndex <> IsVoid And Playlist.colMediaIDs.Count > PlaylistPosition - 1 And IsCur Then
                PlayMedium_ByPlaylistPosition(Playlist.CurMediumIndex + 1)
            Else
                Playlist.IsForwarding = False
            End If
        End If

        If Not DontUpdate Then
            'If IsCur Then
            UpdateGUI(True) ' Alles aktualisieren
            'Else
            '    UpdateGUI(False, False, True)
            'End If
        End If
    End Sub

    ' Gibt die aktuelle Abspielposition in Sekunden zurück.
    ' Gibt IsVoid zurück, wenn die Abspielposition nicht ermittelt werden konnte.
    Public Function GetPlayPos() As Single
        If Playlist.CurMediumID = IsVoid Then Return IsVoid

        Dim Medium As sMedium = Media(Playlist.CurMediumID)
        Select Case Medium.MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                Return afi.GetSoundPlayPos()
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                Return IsVoid
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
            Case Else
                Return IsVoid
        End Select
    End Function

    ' Legt die aktuelle Abspielposition in Sekunden fest.
    ' Gibt IsVoid zurück, wenn die Abspielposition nicht geändert werden konnte.
    Public Function SetPlayPos(ByVal Position As Single) As Single
        If Playlist.CurMediumID = IsVoid Then Return IsVoid

        Dim Medium As sMedium = Media(Playlist.CurMediumID)
        Select Case Medium.MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                Return afi.SetSoundPlayPos(Position)
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                Return IsVoid
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
            Case Else
                Return IsVoid
        End Select
    End Function

    ' Springt einen Schritt in der Abspielposition zurück.
    ' Geht nicht vor die Position "0 Sekunden" zurück.
    ' Die Abspielposition von Netstreams kann nicht verändert werden.
    Public Function StepBack() As Boolean
        If Playlist.CurMediumID = IsVoid Then Return False
        If Media(Playlist.CurMediumID).MediumType = eMediumTypes.eNetstream Then Return False ' Die Abspielposition von Netstreams kann nicht verändert werden.

        Dim Pos As Single, Length As Single
        Pos = GetPlayPos()
        If Pos = IsVoid Then Return False
        Length = GetMediumPlayLen(Playlist.CurMediumID)
        If Length = IsVoid Or Length = 0 Then Return False

        Pos -= PosStep
        If Pos < 0 Then Pos = 0

        Return SetPlayPos(Pos)
    End Function

    ' Springt einen Schritt in der Abspielposition zurück.
    ' Geht nicht über die Position "Abspiellänge des Mediums" hinaus.
    ' Die Abspielposition von Netstreams kann nicht verändert werden.
    Public Function StepForward() As Boolean
        If Playlist.CurMediumID = IsVoid Then Return False
        If Media(Playlist.CurMediumID).MediumType = eMediumTypes.eNetstream Then Return False ' Die Abspielposition von Netstreams kann nicht verändert werden.

        Dim Pos As Single, Length As Single
        Pos = GetPlayPos()
        If Pos = IsVoid Then Return False
        Length = GetMediumPlayLen(Playlist.CurMediumID)
        If Length = IsVoid Or Length = 0 Then Return False

        Pos += PosStep
        If Pos > Length Then Pos = Length

        Return SetPlayPos(Pos)
    End Function

    ' Gibt die Abspiellänge eines Mediums in Sekunden zurück.
    ' Gibt IsVoid zurück, wenn die Abspiellänge nicht ermittelt werden konnte.
    Public Function GetMediumPlayLen(ByVal MediumID As Integer) As Single
        If MediumID < 0 Or MediumID > Media.Count Or Media.Count = 0 Then
            Err.Raise(514, , "Invalid MediumID: " + MediumID.ToString)
            Exit Function
        End If

        Select Case Media(MediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream  ' Netstreams haben evtl. keine Länge
                Return afi.GetSoundPlayLen(MediumID)
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                Return IsVoid
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
            Case Else
                Return IsVoid
        End Select
    End Function

    ' Gibt die Abspiellänge des AKTUELLEN Mediums in Sekunden zurück.
    ' Gibt IsVoid zurück, wenn die Abspiellänge nicht ermittelt werden konnte.
    Public Function GetPlayLen() As Single
        If Playlist.CurMediumID = IsVoid Then Return IsVoid

        Return GetMediumPlayLen(Playlist.CurMediumID)
    End Function

    ' Pausiert die Wiedergabe eines Mediums
    Public Sub PausePlaying()
        If Playlist.CurMediumID = IsVoid Then Exit Sub

        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                afi.PauseSound()
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
            Case Else
                ' ...
        End Select
    End Sub

    ' Gibt an, ob das akt. Medium pausiert ist.
    ' Gibt false zurück, falls das nicht festgestellt werden konnte.
    Public Function IsPaused() As Boolean
        If Playlist.CurMediumID = IsVoid Then Return False

        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                Return afi.IsSoundPaused()
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                Return False
            Case Else
                ' ...
                Return False
        End Select
    End Function

    ' Gibt an, ob gerade ein Medium wiedergegeben wird.
    ' ACHTUNG: Wenn das akt. Medium pausiert ist, wird ebenfalls "true" zurück gegeben!!!
    ' Daher muss ggf. ebenfalls IsReallyPlaying() abgefragt werden.
    Public Function IsPlaying() As Boolean
        If Playlist.CurMediumID = IsVoid Then Exit Function

        Select Case Media(Playlist.CurMediumID).MediumType
            Case eMediumTypes.eSoundfile, eMediumTypes.eNetstream
                Return afi.IsSoundPlaying()
            Case eMediumTypes.eVideo
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
                Return False
                ' Platzhalter für zukünftige Funktionalität
                ' Platzhalter für zukünftige Funktionalität
            Case Else
                Return False
        End Select
    End Function

    ' Gibt an, ob gerade ein Medium wirklich wiedergegeben wird.
    ' D.h., wenn das Medium pausiert ist, wird "false" zurückgegeben.
    Public Function IsReallyPlaying() As Boolean
        Return CBool(IsPlaying() Xor IsPaused())
    End Function

    ' Gibt an, ob ein Dateiname zu einem bekannten Format für eine Mediendatei zählt.
    ' Playlisten gehören NICHT dazu!
    Public Function IsMediaFile(ByVal Filename As String) As Boolean
        If IsInStringList(SoundFilenameExtensions, GetFilenameExtension(Filename)) Then Return True
        ' Hier ggf. Code für andere Formate (Videos) einfügen.
        Return False
    End Function

    ' Lädt die Programmeisntellungen.
    Public Function LoadSettings() As Boolean
        Dim sr As IO.StreamReader
        Dim Line, Element, Value As String
        Dim Pos As Integer

        Try
            sr = New IO.StreamReader(AppPath + "\" + SettingsFile)

            Do
                Line = sr.ReadLine()
                If IsNothing(Line) Then Exit Do

                Line = Trim(Line)
                If Not (Line.StartsWith("'") Or Line.StartsWith("[") Or Line = "") Then
                    Pos = InStr(Line, "=")
                    If Pos > 0 Then
                        Element = Left(Line, Pos - 1).ToUpper
                        Value = Mid(Line, Pos + 1)
                        Try
                            Select Case Element
                                Case "USERVOLUME"
                                    If Not (Val(Value) < 0 Or Val(Value) > 100) Then UserVolume = Val(Value)
                                Case "XPOS"
                                    If Val(Value) < 0 Then Value = "0"
                                    If Val(Value) > Screen.PrimaryScreen.Bounds.Width Then Value = Screen.PrimaryScreen.Bounds.Width.ToString
                                    frmAudiaStartupLocation.X = Val(Value)
                                Case "YPOS"
                                    If Val(Value) < 0 Then Value = "0"
                                    If Val(Value) > Screen.PrimaryScreen.Bounds.Height - 1 Then Value = Screen.PrimaryScreen.Bounds.Height - 1
                                    frmAudiaStartupLocation.Y = Val(Value)
                                Case "DEFAULTQUICKPLAYURL"
                                    DefaultQuickplayURL = Value
                                Case "REPEATPLAYLIST"
                                    RepeatPlaylist = CBool(Value)
                                Case "SKINNAME"
                                    If IO.Directory.Exists(AppPath + "\" + SkinsDir + "\" + Value) Then Skinname = Value
                                Case "CURPLAYLISTINDEX"
                                    If Not (Val(Value) = IsVoid Or Val(Value) < 0 Or Val(Value) > Playlist.colMediaIDs.Count - 1 Or Playlist.colMediaIDs.Count = 0) Then
                                        Playlist.CurMediumIndex = Val(Value)
                                        Playlist.CurMediumID = Playlist.colMediaIDs(Playlist.CurMediumIndex)
                                    End If
                                Case "ENABLECONSOLE"
                                    EnableConsole = CBool(Value)
                                    frmMenu.menConsole.Visible = EnableConsole
                                Case "WINDOWOPACITY"
                                    WindowOpacity = CSng(Val(Value)) ' Val berücksichtigt keine Ländereinstellungen und vermeidet somit Probleme bei untersch. Ländereinstellungen
                                Case "TOPMOST"
                                    AlwaysTopMost = CBool(Value)
                                Case "LANGUAGE"
                                    Language = Value
                                Case "ALLOWAPPUPDATECHECK"
                                    AllowAppUpdateCheck = CBool(Value)
                                Case ""
                                    ' Fehlerhafte Zeile
                                    Log("LoadSettings", Loc("Fehler: Fehlerhafte Zeile beim Auslesen der Programmeinstellungen.") + " (" + Line + ")", True)
                                Case Else
                                    ' Unbekanntes Element
                                    Log("LoadSettings", Loc("Warnung: Unbekannte Bezeichnung einer Programmeinstellung.") + " (" + Element + ")", False)
                            End Select
                        Catch ex As Exception
                            Log("LoadSettings", Loc("Fehler: Eine Programmeinstellung konnte nicht korrekt ausgelesen werden.") + " (" + Element + ")", True)
                        End Try
                    End If
                End If
            Loop

            sr.Close()
            Return True
        Catch ex As Exception
            Log("LoadSettings", Loc("Fehler: Programmeinstellungen konnten nicht korrekt geladen werden.") + " (" + SettingsFile + ")", True)
            MsgBox(Loc("Fehler: Die Programmeinstellungen konnten nicht korrekt ausgelesen werden. Möglicherweise wurden Einstellungen nicht (richtig) geladen."), MsgBoxStyle.Exclamation, Loc("Fehler"))
            Return False
        End Try
    End Function

    ' Speichert die Programmeinstellungen.
    Public Function SaveSettings() As Boolean
        Dim sw As IO.StreamWriter

        Try
            sw = New IO.StreamWriter(AppPath + "\" + SettingsFile)

            sw.WriteLine("' Setting file.")
            sw.WriteLine("'  ")
            sw.WriteLine("' Please do not change any settings!")
            sw.WriteLine("' Please do not edit this file!")
            sw.WriteLine("'  ")
            sw.WriteLine("[Global]")
            sw.WriteLine("UserVolume=" + UserVolume.ToString)
            sw.WriteLine("XPos=" + frmAudia.Location.X.ToString)
            sw.WriteLine("YPos=" + frmAudia.Location.Y.ToString)
            sw.WriteLine("DefaultQuickplayURL=" + DefaultQuickplayURL)
            sw.WriteLine("Skinname=" + Skinname)
            sw.WriteLine("RepeatPlaylist=" + RepeatPlaylist.ToString)
            sw.WriteLine("CurPlaylistIndex=" + Playlist.CurMediumIndex.ToString)
            sw.WriteLine("EnableConsole=" + EnableConsole.ToString)
            sw.WriteLine("WindowOpacity=" + Str(WindowOpacity)) ' Str() berücksichtigt keine Ländereinstellungen und vermeidet somit Probleme bei untersch. Ländereinstellungen
            sw.WriteLine("TopMost=" + AlwaysTopMost.ToString)
            sw.WriteLine("Language=" + Language)
            sw.WriteLine("AllowAppUpdateCheck=" + AllowAppUpdateCheck.ToString)

            sw.Close()
            Return True
        Catch ex As Exception
            Log("SaveSettings", Loc("Fehler: Programmeisntellungen konnten nicht korrekt gespeichert werden.") + " (" + SettingsFile + ")", True)
            MsgBox(Loc("Fehler: Die Programmeinstellungen konnten nicht korrekt ausgelesen werden. Möglicherweise wurden Einstellungen nicht (richtig) geladen."), MsgBoxStyle.Exclamation, Loc("Fehler"))
            Return False
        End Try
    End Function

    ' Zeigt eine Notify-Einblendung an.
    Public Sub ShowNotify(ByVal TitleText As String, Optional ByVal NotifyTime As Single = 5)
        frmNotifier.Notify(TitleText, NotifyTime)
    End Sub

    ' Lokalisiert einen String.
    Public Function Loc(ByVal NativeString As String, Optional ByVal ID As Integer = IsVoid) As String
        ' Wichtige Konstante: DefaultLanguage
        Return NativeString
    End Function

    ' Tätigt einen (lokalisierten) Eintrag in die Protokoll-Datei
    ' Beispiel für Sender: "Funktion TestFunktion" (ohne Klammern hinter dem Funktionsname!)
    Public Sub Log(Optional ByVal Sender As String = "Global", Optional ByVal LogMessage As String = "", Optional ByVal IsErrorMsg As Boolean = False)
        Dim Adds As String = ""
        Dim LogLines As String = ""
        Dim NowStarted As Boolean

        If LogFile = "" Then Exit Sub
        If LogServiceStarted = False Then
            Try
                Dim fi As System.IO.FileInfo = My.Computer.FileSystem.GetFileInfo(AppPath + "\" + LogFile)
                If fi.Length > MaxLogFileSize Then ' Length() liefert die Dateigröße in Bytes zurück
                    My.Computer.FileSystem.DeleteFile(AppPath + "\" + LogFile) ' Die Log-Datei löschen (leeren), wenn sie zu groß geworden ist (festgelegt durch MaxLogFileSize)
                End If
            Catch ex As Exception
                ' Fehlerbehandlung: keine. Die Datei kann ja fehlen, weil sie zuvor zu groß war und gelöscht wurde.
            End Try
            LogServiceStarted = True
            NowStarted = True ' Gibt an, ob der Protokollservice bei genau diesem Funktionsaufruf gestartet wurde (Nutzen: Siehe weiter unten)
        End If

        Try
            Dim sw As New IO.StreamWriter(AppPath + "\" + LogFile, True)
            Dim Line As String

            If IsErrorMsg = True Then Adds = "!! " Else Adds = "   " ' Zusatz, der am Zeilenbeginn eingefügt wird, um Fehlermeldungen von Standard-Protokollnachrichten zu trennen.
            If NowStarted = True Then ' Beim Start des Protokollservice:
                sw.WriteLine("")  ' Leerzeilen einfügen - dient zur optischen Abgrenzung einzelner Protokolle
                sw.WriteLine("")
                sw.WriteLine("____________________________________________________________________________________") ' horiz. Trennstrich einfügen
                sw.WriteLine("")
            End If
            Line = Adds + FormatDateTime(Now.Date.ToString, DateFormat.ShortDate) + " " + FormatDateTime(Now.TimeOfDay.ToString, DateFormat.ShortTime) + ":" + (Now.Second + (Now.Millisecond / 1000)).ToString + " [" + Sender + "]: " + LogMessage ' Message generieren
            sw.WriteLine(Line) ' Message schreiben
            If EnableConsole Then frmConsole.txtGlobalLog.AppendText(Line + vbCrLf)
            Debug.Print(Line)

            sw.Close()
        Catch ex As Exception
            ' mögliche Fehlerquelle: Festplatte voll belegt
            ' Fehlerbehandlung: keine. Dieser Fehler besitzt eine geringe Priorität. Keine Benachrichtung des Benutzers nötig.
        End Try
    End Sub

    ' Dieser Timer kümmert sich um periodische Vorgänge, die nur "selten" durchgeführt werden müssen und eine geringe Priorität haben.
    Private Sub LowLevelTimer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles LowLevelTimer.Tick
        PortablesDetection()
    End Sub

    ' Zeigt den Ordner an, in dem eine Medientdatei gespeichert ist, und selektiert die Mediendatei.
    Public Function ShowMediumStoragePlace(Optional ByVal MediumID As Integer = IsVoid) As Boolean
        If MediumID = IsVoid Then MediumID = Playlist.CurMediumID
        If MediumID <> IsVoid Then
            If IsURL(Media(MediumID).Filename) = False Then
                Diagnostics.Process.Start(GetPath(Media(MediumID).Filename)) ' Ordner öffnen/anzeigen
                Threading.Thread.Sleep(350) ' 350 Millisek. warten, damit der Ordner auf jeden Fall geöffnet worden ist.
                SendKeys.Send(Media(MediumID).FileTitle) ' Den Dateititel (z.B. "Chihiro - Infection.mp3") als Tastatureingabe senden. Dadurch wird die Datei markiert!
                Return True
            End If
        End If
        Return False
    End Function

    ' Durchsucht die virtuelle Medienbiliothek.
    ' Gibt die Anzahl der gefundenen Elemente zurück.
    Public Function SearchLibrary(ByVal SearchString As String) As Integer
        If SearchString = "" Then Return IsVoid

        Dim i As Integer, f As Integer = 0
        For i = 1 To Media.Count - 1
            If strContains(Media(i).FileTitle, SearchString, CompareMethod.Text) Or strContains(Media(i).Title, SearchString, CompareMethod.Text) Or strContains(Media(i).Artist, SearchString, CompareMethod.Text) Then
                AddToPlaylist(i)
                f += 1
            End If
        Next
        UpdateGUI(False, False, True)
        Return f
    End Function

    ' Überprüft, ob ein "globaler" Tastendruck erfolgt ist.
    ' D.h. es werden auch dann Tastendrücke ausgewertet, wenn das Programm nicht den Fokus hat oder minimiert ist.
    Public Sub CheckGlobalKeys()
        For i = 0 To KeyStates.Count - 1
            Dim KeyState As sKeyState = KeyStates(i)
            If GetKeyState(KeyState.KeyCode) = -127 Or GetKeyState(KeyState.KeyCode) = -128 Then ' Taste ist gedrückt:
                If (Now - KeyState.TimeWhenPressed).TotalMilliseconds >= 300 Then
                    KeyState.TimeWhenPressed = Now
                    KeyStates(i) = KeyState

                    Select Case KeyState.KeyCode
                        Case 176  ' 176 = Erweiterte Taste "Vor", z.B. auf Logitech Elite Keyboard
                            PlaylistForward()
                        Case 177 ' 177 = Erweiterte Taste "Zurück", z.B. auf Logitech Elite Keyboard
                            PlaylistBack()
                        Case 178 ' 178 = Erweiterte Taste "Stop", z.B. auf Logitech Elite Keyboard
                            StopPlaying()
                        Case 179  ' 179 = Erweiterte Taste "Play/Pause", z.B. auf Logitech Elite Keyboard
                            If Playlist.CurMediumIndex <> IsVoid Then
                                If IsReallyPlaying() Then
                                    PausePlaying()
                                Else
                                    PlayMedium_ByPlaylistPosition(Playlist.CurMediumIndex + 1)
                                End If
                            End If
                    End Select
                End If
            End If
        Next

        ' DEBUG-Code zum Testen, ob irgendeine Taste gedrückt wird:
        ' (Dadurch kann dann der Keycode einer bestimmten Taste herausgefunden werden!)
        'For i = 0 To 200
        '    If GetKeyState(i) = -128 Or GetKeyState(i) = -127 Then Debug.Print(i)
        'Next
    End Sub

    ' (Diese Fukntion fast nur mehrere Befehlsaufrufe zusammen)
    Public Sub SetGlobalKeys()
        AddGlobalKey(176)
        AddGlobalKey(177)
        AddGlobalKey(178)
        AddGlobalKey(179)
    End Sub

    ' Fügt eine Taste zur Liste der zu überwachenden Tasten hinzu.
    ' Die Taste wird über den "normalen" VB-Keycode identifiziert.
    Public Sub AddGlobalKey(ByVal Keycode As Long)
        Dim KeyState As New sKeyState
        KeyState.KeyCode = Keycode
        KeyState.TimeWhenPressed = Now
        KeyStates.Add(KeyState)
    End Sub

    ' Fügt das aktuelle Medium den Favoriten hinzu.
    ' Der Titel kann dabei angegeben werden.
    Public Sub AddCurToFavorits()
        If Playlist.CurMediumID = IsVoid Then Exit Sub

        Dim Title As String
        Title = Trim(InputBox(Loc("Bitte geben sie den gewünschten Titel an." + vbCr + vbCr + Loc("Es muss ein Titel angegeben werden.")), Loc("Titel"), Media(Playlist.CurMediumID).GetBestTitle))
        If Title = "" Then Exit Sub

        Dim Favorit As New sFavorit
        Favorit.Title = Title
        Favorit.Target = Media(Playlist.CurMediumID).Filename
        Favorits.Add(Favorit)

        UpdateFavoritsMenu()
        frmFavorits.Update_lvFavorits()
    End Sub

    ' Übergeordente Fukntion: Prüft, ob es ein neues Programmupdate (via Internet) gibt.
    ' Dient nur dazu, den eigentlichen Update-Prozess als einen Backgroundworker-Prozess auszuführen
    Public Sub AppUpdateCheck()
        If AllowAppUpdateCheck Then
            Dim BW As New System.ComponentModel.BackgroundWorker

            AddHandler BW.DoWork, AddressOf DownloadAppUpdateInfo
            BW.RunWorkerAsync()
        End If
    End Sub

    ' Lädt eine Datei herunter, die Updateinformationen enthält.
    Public Sub DownloadAppUpdateInfo()
        Try
            If My.Computer.FileSystem.FileExists(AppPath + "\updateinfo.tmp") Then My.Computer.FileSystem.DeleteFile(AppPath + "\updateinfo.tmp")

            My.Computer.Network.DownloadFile(AppUpdateURL, AppPath + "\updateinfo.tmp")

            Try
                Dim sr As IO.StreamReader
                Dim UpdateVersion As String, AppUpdateFileURL As String
                Dim UpdateVersionBuild As Integer

                sr = New IO.StreamReader(AppPath + "\updateinfo.tmp")
                UpdateVersionBuild = Val(sr.ReadLine())
                UpdateVersion = sr.ReadLine()
                AppUpdateFileURL = sr.ReadLine()
                sr.Close()
                My.Computer.FileSystem.DeleteFile(AppPath + "\updateinfo.tmp")

                If UpdateVersionBuild > 0 Then ' Sicherheitsabfrage
                    If UpdateVersionBuild > VersionBuild Then ' Es gibt eine neuere Version!!!
                        Log(, Loc("Update-Überprüfung: Es ist ein neues Update verfügbar.") + " (" + Version + " -> " + UpdateVersion + ")")
                        If MsgBox(Loc("Es ist ein Update auf eine neue Version verfügbar.") + vbCr + vbCr + Loc("Ihre Version: ") + Version + vbCr + Loc("Verfügbare Version: ") + UpdateVersion + vbCr + vbCr + Loc("Soll das Update heruntergeladen werden?"), MsgBoxStyle.Information + MsgBoxStyle.YesNo, Loc("Update verfügbar")) = MsgBoxResult.Yes Then
                            Log(, Loc("Update-Überprüfung: Versuche, Update herunter zu laden.") + " (" + UpdateVersion + ")")

                            Dim BW As New System.ComponentModel.BackgroundWorker

                            AddHandler BW.DoWork, AddressOf DownloadAppUpdate
                            RealAppUpdateURL = AppUpdateFileURL
                            BW.RunWorkerAsync()
                        End If ' Ansonsten Abbruch.
                    Else
                        Log(, Loc("Update-Überprüfung: Programm ist auf dem neuesten Stand.") + " (" + Version + ")")
                    End If
                End If
            Catch ex As Exception
                Exit Sub ' ABBRUCH! (Aus Sicherheitsgründen darf nicht versucht werden, das Update fortzusetzen!)
            End Try
        Catch ex As Exception
            Exit Sub ' ABBRUCH! WICHTIG! Es darf auf keinen Fall versucht werden, das Update fortzusetzen!
        End Try
    End Sub

    ' Lädt die neueste Programmversion herunter.
    Public Sub DownloadAppUpdate()
        Try
            Dim Target As String = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory)
            Dim Pos As Integer = InStrRev(RealAppUpdateURL, "/")

            If Pos = 0 Then Exit Sub ' ABBRUCH!
            Target += "\" + Mid(RealAppUpdateURL, Pos + 1)

            If My.Computer.FileSystem.FileExists(Target) Then My.Computer.FileSystem.DeleteFile(Target)

            My.Computer.Network.DownloadFile(RealAppUpdateURL, Target)

            MsgBox(Loc("Das Update wurde erfolgreich heruntergeladen und unter folgendem Ziel gespeichert:") + vbCr + vbCr + Target + vbCr + vbCr + Loc("Bitte beenden Sie") + " " + Application.ProductName + " " + Loc("und führen Sie das Update aus."), MsgBoxStyle.Information, Loc("Update herunter geladen"))
            Log(, Loc("Update wurde erfolgreich herunter geladen"))
        Catch ex As Exception
            Exit Sub ' ABBRUCH! WICHTIG! Es darf auf keinen Fall versucht werden, das Update fortzusetzen!
        End Try
    End Sub
End Module
