﻿Imports FMOD
'
' afi = Audia FMod-Interface
' Dieses Modul bildet die  Schnittstelle zwischen Audia und der FMod-Engine.
' Es ist zuständig für die Wiedergabe von Sounddateien und Netstreams.
'
Module afi
    ' Required for GetStringFromPointer
    Public Declare Function lstrcpy Lib "kernel32" Alias "lstrcpyA" (ByVal lpString1 As String, ByVal lpString2 As Long) As Long
    Public Declare Function lstrlen Lib "kernel32" Alias "lstrlenA" (ByVal lpString As Long) As Long

    Private FRESULT As New FMOD.RESULT
    Private FSystem As New FMOD.System
    Private Channel As FMOD.Channel

    Public Const StreamBufferSize = 65536 ' Angabe in Bytes, wichtig für NetStreams

    ' Initialisiert die FMod-Engine.
    Public Sub Start()
        FRESULT = FMOD.Factory.System_Create(FSystem)
        CheckFModError(FRESULT)

        FRESULT = FSystem.setOutput(OUTPUTTYPE.AUTODETECT)
        CheckFModError(FRESULT)

        FRESULT = FSystem.init(100, FMOD.INITFLAG.NORMAL, Nothing)
        CheckFModError(FRESULT)

        FRESULT = FSystem.setStreamBufferSize(StreamBufferSize, TIMEUNIT.PCMBYTES) ' Die File-Buffer-Größe für die Wiedergabe von Net-Streams etwas erhöhen, um Lags zu vermeiden. Die Größe des Decode-Buffers wird dadurch nicht verändert.
        CheckFModError(FRESULT)
    End Sub

    ' Beendet die FMod-Engine.
    Public Sub Shutdown()
        FRESULT = FSystem.close
        CheckFModError(FRESULT)
    End Sub

    ' Fehler-Überprüfung. Falls ein Fehler aufgetreten ist, wird eine (englischsprachige) Fehlermeldung als String zurückgegeben.
    Private Function CheckFModError(ByVal myFresult As FMOD.RESULT) As String
        If myFresult <> FMOD.RESULT.OK Then
            Log("FMod Engine", "FMod error: " + myFresult.ToString + " - " + FMOD.Error.String(myFresult))
            Debug.Print("FMOD error: " + myFresult.ToString + " - " + FMOD.Error.String(myFresult))
            Return FMOD.Error.String(myFresult)
        Else
            Return ""
        End If
    End Function

    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    ' EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! EXPERIMENTELL! 
    Public Function AddEqualizer() As Boolean
        Static FirstRun As Boolean = True

        If FirstRun = False Then Exit Function
        FirstRun = True

        'Dim myDSP As FMOD.DSP = Nothing
        'FRESULT = FSystem.createDSPByType(DSP_TYPE.PARAMEQ, myDSP)
        'FRESULT = Channel.addDSP(myDSP)
        'FRESULT = myDSP.setParameter(FMOD.DSP_PARAMEQ.CENTER, 8000)
        'FRESULT = myDSP.setParameter(FMOD.DSP_PARAMEQ.BANDWIDTH, 1)
        'FRESULT = myDSP.setParameter(FMOD.DSP_PARAMEQ.BANDWIDTH, 2)
    End Function

    ' Lädt, sofern nötig, einen Sound und spielt diesen ab.
    ' Gibt "true" zurück, wenn das Laden/Abspieln erfolgreich war
    ' Diese Funktion sollte indirekt über PlayMedium() aufgerufen werden.
    ' ForceNew ezwingt den "Neustart" des Sounds, auch wenn er pausiert ist.
    Public Function PlaySound(ByVal MediumID As Long, Optional ByVal ForceNew As Boolean = False) As Boolean
        Dim Medium As sMedium

        If MediumID < 0 Or MediumID > Media.Count Or Media.Count = 0 Then
            Err.Raise(514, , "Invalid MediumID: " + MediumID.ToString)
            Exit Function
        End If
        Medium = Media(MediumID)

        If IsNothing(Medium.FModSound) Or (Medium.MediumType = eMediumTypes.eNetstream And IsSoundPaused() = False) Then ' Netstreams müssen immer neu geladen werden, wenn ihre Wiedergabe unterbrochen wurde - außer, sie wurden pausiert.
            Dim FModSound As New FMOD.Sound
            If Medium.MediumType = eMediumTypes.eNetstream Or Medium.MediumType = eMediumTypes.eUndefined Then ShowNotify(Loc("Buffering..."))
            FRESULT = FSystem.createStream(Medium.Filename, (FMOD.MODE._2D Or FMOD.MODE.SOFTWARE Or FMOD.MODE.CREATESTREAM), FModSound) ' Neuen Stream erzeugen (Datei/Netstream laden). "SOFTWARE" wird für Equalizer-Funktionalität benötigt.
            If FRESULT <> FMOD.RESULT.OK Then
                Medium.MediumType = eMediumTypes.eError ' Das Format bzw. die Datei kann offensichtlich nicht wiedergegeben werden.
                Media(MediumID) = Medium
                CheckFModError(FRESULT)
                Return False ' Fehler beim Laden 
            Else
                Medium.FModSound = FModSound
                If Medium.MediumType = eMediumTypes.eUndefined Then
                    Medium.MediumType = eMediumTypes.eSoundfile ' Das Format kann offensichtlich wiedergegeben werden, daher wird es angenommen.
                End If
                Media(MediumID) = Medium
            End If
        End If

        GetSoundTags(MediumID) ' Sound-Tags ermitteln (Titel, Interpret usw)

        If ForceNew Or IsSoundPaused() = False Then
            StopSound() ' Stoppen -> Sound wird wieder von vorne abgespielt
            FRESULT = FSystem.playSound(FMOD.CHANNELINDEX.FREE, Medium.FModSound, False, Channel)
            If FRESULT <> FMOD.RESULT.OK Then
                CheckFModError(FRESULT)
                Return False ' Abspielen ist nicht möglich
            End If
        Else
            PauseSound(False)
        End If

        SetSoundVolume(UserVolume) ' Notwendig. So kann der User bereits eine Lautstärke festlegen, ehe ein FModSound geladen wurde. Das ist per SetVolume() nicht im Vorasu möglich.

        Return True
    End Function

    ' Stoppt die Wiedergabe des akt. Mediums (geht an den Anfang des Mediums zurück)
    Public Sub StopSound()
        If IsSoundPlaying() = True Then
            Channel.stop() ' Hinweis: Überprüfung auf IsNothing(Channel) wird schon vorher durchgeführt.
            Channel = Nothing ' Notwendig. Ein gestoppter Channel wird freigegeben und kann nicht noch einmal verwendet werden.
        End If
    End Sub

    ' Prüft, ob gerade ein Medium wiedergegeben wird.
    ' Gibt false zurück, wenn das nicht festgestellt werden konnte.
    Public Function IsSoundPlaying() As Boolean
        Dim IsCurrentlyPlaying As Boolean

        If IsNothing(Channel) Then Return False
        Channel.isPlaying(IsCurrentlyPlaying) ' Prüfen, ob ein Medium abgespielt wird
        Return IsCurrentlyPlaying
    End Function

    ' Gibt die Abspielposition des aktuellen Audio-Mediums in Sekunden zurück.
    ' Wenn ein FMod-Sound noch nicht geladen wurde, wird dies nachgeholt.
    ' Gibt die Konstante IsVoid zurück, falls die Länge nicht bestimmt werden kann.
    ' Dies trifft auch für NetStreams zu. (NetStreams haben keine Länge!)
    Public Function GetSoundPlayLen(ByVal MediumID As Long) As Single
        Dim Medium As sMedium
        Dim MilliSecs As Integer

        If MediumID < 0 Or MediumID > Media.Count Or Media.Count = 0 Then
            Err.Raise(514, , "Invalid MediumID: " + MediumID.ToString)
            Exit Function
        End If
        Medium = Media(MediumID)

        If Medium.MediumType = eMediumTypes.eError Then Return IsVoid
        If Medium.MediumType = eMediumTypes.eNetstream And Medium.HasPlayLen = False And Medium.IsPlayLenTested Then Return False ' NetStream wurde darauf getestet, ob er eine Länge hat, er hat jedoch keine.

        If IsNothing(Medium.FModSound) Then ' Sound laden, falls noch nicht geschehen (notwendig!)
            If Medium.MediumType = eMediumTypes.eNetstream Then Return IsVoid ' Das Laden von NetStreams wäre möglich; hier aber ungünstig (aufgrund der evtl. langen Ladezeit)

            Dim FModSound As New FMOD.Sound
            FRESULT = FSystem.createStream(Medium.Filename, (FMOD.MODE._2D Or FMOD.MODE.HARDWARE Or FMOD.MODE.CREATESTREAM), FModSound) ' Neuen Stream erzeugen (Datei/Netstream laden)
            If FRESULT <> FMOD.RESULT.OK Then
                CheckFModError(FRESULT)
                Medium.MediumType = eMediumTypes.eError ' Vllt. Datei nicht mehr existent oder beschädigt etc.
                Media(MediumID) = Medium
                Return IsVoid
            Else
                Medium.FModSound = FModSound
                If Medium.MediumType = eMediumTypes.eUndefined Then
                    Medium.MediumType = eMediumTypes.eSoundfile ' Das Format kann offensichtlich wiedergegeben werden, daher wird es angenommen.
                End If
                Media(MediumID) = Medium
            End If
        End If

        Try
            FRESULT = Medium.FModSound.getLength(MilliSecs, TIMEUNIT.MS)
            If FRESULT = RESULT.OK Then
                If Medium.MediumType = eMediumTypes.eNetstream Then
                    Medium.HasPlayLen = True
                    Medium.IsPlayLenTested = True
                    Media(MediumID) = Medium
                End If
                Return CSng(MilliSecs / 1000) ' Umrechnung in Sekunden
            Else
                Return IsVoid
            End If
        Catch ex As Exception ' Beim Abfragen der Länge eines Netstreams tritt ein Fehler auf. Wenn der Medientyp unbekannt ist, kann es aber sein, dass die Länge eines Netstreams abgefragt werden soll.
            If Medium.MediumType = eMediumTypes.eNetstream Then
                Medium.HasPlayLen = False
                Medium.IsPlayLenTested = True
                Media(MediumID) = Medium
            End If
            Return IsVoid
        End Try
    End Function

    ' Legt die Abspielposition des aktuellen Audio-Mediums fest - Angabe in Sekunden
    ' Gibt true zurück, wenn dies erfolgreich geschehen ist.
    Public Function SetSoundPlayPos(ByVal Position As Single) As Boolean
        If IsNothing(Channel) Then Return False
        FRESULT = Channel.setPosition(Position * 1000, TIMEUNIT.MS) ' Legt die Abspielposition in Millisekunden fest
        Return (FRESULT = RESULT.OK)
    End Function

    ' Gibt die Abspielposition des aktuellen Audio-Mediums in Sekunden zurück.
    ' ACHTUNG: Gibt 0 zurück, wenn kein FModSound geladen wurde.
    Public Function GetSoundPlayPos() As Single
        Dim MilliSecs As Long

        If IsNothing(Channel) Then Return 0
        FRESULT = Channel.getPosition(MilliSecs, TIMEUNIT.MS)  ' Ermittelt die Abspielposition in Millisekunden
        CheckFModError(FRESULT)

        Return CSng(MilliSecs / 1000) ' Umrechnung in Sekunden
    End Function

    ' Legt die Lautstärke der Tonausgabe von Audio-Medien fest
    ' Volume: 0-100 %
    Public Function SetSoundVolume(ByVal Volume As Single) As Boolean
        If IsNothing(Channel) Then Return False
        FRESULT = Channel.setVolume(Volume / 100) ' Umrechnung aus Prozentwert
        Return CBool(FRESULT = RESULT.OK)
    End Function

    ' Fragt die Lautstärke der Tonausgabe von Audio-Medien ab
    ' Volume: 0-100 %
    ' ACHTUNG: Gibt auch dann 0 zurück, wenn kein FModSound geladen wurde. (Nach Laden eines Sounds wird die Lautstärke dann bei 100% stehen.)
    Public Function GetSoundVolume() As Single
        Dim Volume As Single

        If IsNothing(Channel) Then Return 0
        FRESULT = Channel.getVolume(Volume)
        Return Volume * 100 ' Umrechnung in Prozentwert
    End Function

    ' Deaktiviert oder aktiviert die Tonausgabe für Audio-Medien-Wiedergabe (diese werden aber unabhängig davon weiter abgespielt)
    ' Liefert True zurück, wenn der Modus geändert werden konnte.
    Public Function SetSoundMuteMode(ByVal Mute As Boolean) As Boolean
        If IsNothing(Channel) Then Return False
        FRESULT = Channel.setMute(Mute)
        Return CBool(FRESULT = FMOD.RESULT.OK)
    End Function

    ' Fragt den Status der Tonausgabe für Audio-Medien-Wiedergabe ab.
    ' ACHTUNG: Gibt auch dann False zurück, wenn kein FModSound geladen wurde.
    Public Function GetSoundMuteMode() As Boolean
        Dim Mute As Boolean
        If IsNothing(Channel) Then Return False
        FRESULT = Channel.getMute(Mute)
        Return Mute
    End Function

    ' Pausiert die Wiedergabe eines Sounds
    Public Sub PauseSound(Optional ByVal Pause As Boolean = True)
        If IsNothing(Channel) Then Exit Sub
        FRESULT = Channel.setPaused(Pause) ' pausieren bzw. Pause beenden
        If Not FRESULT = RESULT.ERR_INVALID_PARAM Then CheckFModError(FRESULT) ' Es treten teilweise Fehlermeldungen auf, die aber nicht wirkliche Fehler sind.
    End Sub

    ' Gibt an, ob der akt. Sound pausiert ist
    ' Gibt false zurück, falls das nicht festgestellt werden konnte.
    Public Function IsSoundPaused()
        Dim Paused As Boolean = False
        If IsNothing(Channel) Then Return False
        FRESULT = Channel.getPaused(Paused)
        If Not FRESULT = RESULT.ERR_INVALID_PARAM Then CheckFModError(FRESULT) ' Es treten teilweise Fehlermeldungen auf, die aber nicht wirkliche Fehler sind.
        Return Paused
    End Function

    ' Ermittelt alle von Audia verwendeten Tags eines Soundmediums.
    ' Die Werte der Tags werden in den Eigenschaften des Mediums gespeichert.
    Public Function GetSoundTags(ByVal MediumID As Integer) As Boolean
        Dim tagsupdated As Long
        Dim Numtags As Long
        Dim nullstring As String ' Wichtig: diesen String nicht initialisieren!
        Dim Updated As Boolean = False

        Dim Medium As sMedium = Media(MediumID)

        If IsNothing(Medium.FModSound) Then Return False
        FRESULT = Medium.FModSound.getNumTags(Numtags, tagsupdated)
        CheckFModError(FRESULT)

        If tagsupdated <> 0 Then
            For i = tagsupdated To 1 Step -1
                Dim Tag As FMOD.TAG ' Wichtig: Nicht als "Nothing" initialisieren!

                FRESULT = Medium.FModSound.getTag(nullstring, -1, Tag)
                If FRESULT <> RESULT.OK Then
                    Exit For
                End If

                If Tag.datatype = TAGDATATYPE.STRING Then
                    Dim Format As FMOD.SOUND_FORMAT
                    Dim Channels As Long
                    Dim Bits As Long
                    Dim tagname As String
                    Dim tagdata As String
                    Dim gSoundType As FMOD.SOUND_TYPE

                    FRESULT = Medium.FModSound.getFormat(gSoundType, Format, Channels, Bits)
                    CheckFModError(FRESULT)

                    tagname = Space(Len(Tag.name))
                    tagname = Tag.name.ToUpper

                    tagdata = Space(Tag.datalen - 1)
                    Call lstrcpy(tagdata, Tag.data)

                    Select Case tagname
                        Case "ARTIST", "TPE1"
                            tagdata = Space(Tag.datalen - 1)
                            Call lstrcpy(tagdata, Tag.data)
                            If tagdata.EndsWith(Chr(0)) Then tagdata = Left(tagdata, tagdata.Length - 1)
                            tagdata = tagdata.Trim
                            Medium.Artist = tagdata
                            Updated = True
                        Case "TITLE", "TIT2", "ICY-NAME"
                            tagdata = Space(Tag.datalen - 1)
                            Call lstrcpy(tagdata, Tag.data)
                            If tagdata.EndsWith(Chr(0)) Then tagdata = Left(tagdata, tagdata.Length - 1)
                            tagdata = tagdata.Trim
                            Medium.Title = tagdata
                            Updated = True
                        Case "ALBUM"
                            ' Code
                        Case "GENRE", "ICY-GENRE"
                            ' Code
                        Case "TYER" ' (Erscheinungs-)Jahr
                            ' Code
                        Case "TRCK" ' Tracknr. (Achtung: Format etwas unklar. Bsp für mögl. Wert: "9/14")
                            ' Code
                        Case "ICY-URL" ' URL des Net-Stream-Broadcaster
                            ' Code
                        Case Else
                            ' Code?
                    End Select
                Else
                    Exit For
                End If
            Next
        End If

        If Updated Then Media(MediumID) = Medium

        Return True
    End Function
End Module
