﻿Module ahelp
    Public Const StandardSeparator = "|"

    ' Gibt an, ob eine Taste gedrückt wurde:
    Public Declare Function GetKeyState Lib "user32" (ByVal nVirtKey As Long) As Integer

    Public Structure sKeyState
        Dim KeyCode As Long
        Dim TimeWhenPressed As Date
    End Structure
    Public KeyStates As New List(Of sKeyState)

    ' Gibt "" zurück, wenn nichts geparst werden konnte
    Public Function ParseMid(ByVal ParseString As String, ByVal StartChar As String, Optional ByVal EndChar As String = "") As String
        Dim MidString As String
        Dim StartPos As Integer, EndPos As Integer

        If EndChar = "" Then EndChar = StartChar
        If StartChar = "" Then Return ParseString

        StartPos = InStr(ParseString, StartChar)
        If StartPos > 0 Then MidString = Mid(ParseString, StartPos + 1) Else Return ""
        EndPos = InStrRev(MidString, EndChar)
        If EndPos = 0 Then Return ""
        Return Left(MidString, EndPos - 1)
    End Function

    ' Bis zum 1. Vorkommen des (von links) gesuchten Chars 
    Public Function ParseLeft(ByVal ParseString As String, ByVal EndChar As String) As String
        Dim EndPos As Integer

        EndPos = InStr(ParseString, EndChar)
        If EndPos > 0 Then Return Left(ParseString, EndPos - 1) Else Return ParseString
    End Function

    ' Ab dem 1. Vorkommen des (von links) gesuchten Chars 
    Public Function ParseRight(ByVal ParseString As String, ByVal StartChar As String) As String
        Dim StartPos As Integer

        StartPos = InStr(ParseString, StartChar)
        If StartPos > 0 Then Return Mid(ParseString, StartPos + 1) Else Return ""
    End Function

    Public Function ParsePart(ByVal ParseString As String, ByVal Number As Integer, Optional ByVal Separator As String = ",") As String
        Dim Array() As String
        Array = Split(ParseString, Separator)
        If Number < 1 Or Number > Array.GetUpperBound(0) Then Return "" Else Return Array(Number - 1)
    End Function

    ' Prüft, ob ein String in einer Stringliste enthalten ist.
    ' Bsp. für Stringlist: "Anton|Karla|Franz", Bsp. für Suchstring: "Karla"
    Public Function IsInStringList(ByVal StringList As String, ByVal SearchString As String, Optional ByVal Separator As String = StandardSeparator, Optional ByVal CompareMethod As CompareMethod = CompareMethod.Text) As Boolean
        If StringList.StartsWith(Separator) = False Then StringList = Separator + StringList
        If StringList.EndsWith(Separator) = False Then StringList = StringList + Separator

        Return (InStr(StringList, Separator + SearchString + Separator, CompareMethod) > 0)
    End Function

    ' Liefert die Dateiendung eines (möglicherweise fiktiven) Dateinamens.
    ' Es kann eine Pfadangabe im Dateinamen enthalten sein.
    ' Bsp: "Test.bmp" -> liefert "bmp"
    ' Bsp: "Test.bmp.txt" -> liefert "txt"
    ' Bsp: "Test." -> liefert ""
    ' Bsp: "Test" -> liefert ""
    ' Bsp: "C:\Dateien\Test.bmp" -> liefert "bmp"
    Public Function GetFilenameExtension(ByVal Filename As String) As String
        Dim Pos As Integer

        If Filename = "" Then Return ""
        Pos = InStrRev(Filename, ".")
        If Pos = 0 Then
            Return ""
        Else
            If Pos < InStrRev(Filename, "\") Then
                Return ""
            Else
                Return Mid(Filename, Pos + 1)
            End If
        End If
    End Function

    ' Ermittelt den "Titel" eines Dateinames.
    ' Bsp: "C:\Dateien\Test.bmp" -> liefert "Test"
    ' URLs bleiben unverändert.
    Public Function GetFileTitle(ByVal Filename As String) As String
        If IsURL(Filename) Then Return Filename
        Return IO.Path.GetFileNameWithoutExtension(Filename) ' Pfadangaben entfernen & Dateieindung entfernen
    End Function

    ' Entscheidet, ob der übergebene String eine URL darstellt.
    Public Function IsURL(ByVal URL As String) As Boolean
        URL = URL.ToUpper
        Return (URL.StartsWith("HTTP://") = True)
    End Function

    ' Entscheidet, ob der übergebene String eine URL darstellt.
    Public Function IsPlaylist(ByVal Filename As String) As Boolean
        Filename = Filename.ToUpper
        Return (IsInStringList(PlaylistFilenameExtensions, GetFilenameExtension(Filename)))
    End Function

    ' Prüft, ob ein String in einem anderen enthalten ist.
    ' (~ vereinfachter Aufruf der InStr()-Funktion)
    Public Function strContains(ByVal SourceString As String, ByVal SearchString As String, Optional ByVal CompareMethod As CompareMethod = CompareMethod.Text) As Boolean
        Return (InStr(SourceString, SearchString, CompareMethod) > 0)
    End Function

    ' Liefert die Pfadangabe eines Dateinamens zurück.
    ' Bsp: "C:\Test\haus.bmp" -> "C:\Test\"
    ' Bei URLs wird "" zurückgegeben.
    ' Der Parameter ReturnFilePart=true bewirkt, dass nicht der Pfad, sondern der Dateitiel + Endung zurückgegeben wird.
    Public Function GetPath(ByVal Filename As String, Optional ByVal ReturnFilePart As Boolean = False) As String
        Dim Pos As Integer

        If Filename = "" Then Return ""
        Pos = InStrRev(Filename, "\")
        If Pos > 0 Then
            If ReturnFilePart = False Then
                Return Left(Filename, Pos)
            Else
                Return Mid(Filename, Pos + 1)
            End If
        Else
            Return ""
        End If
    End Function

    ' Wandelt einen relativen Pfad in einem absoluten Pfad um
    ' Beispiel: "..\Caendoo\Daewroo.mp3" --> "C:\Musik\Caendoo\Daewroo.mp3"
    ' Achtung: Der Pfad muss absolut angegeben werden und existieren!
    ' URLS werden unverändert zurückggegeben!
    Public Function GetAbsolutFilename(ByVal Path As String, ByVal Filename As String) As String
        If IsURL(Filename) Then Return Filename
        If Filename.StartsWith("\") Then Filename = Mid(Filename, 2) ' anführendes Backslash entfernen

        IO.Directory.SetCurrentDirectory(Path)
        Return IO.Path.GetFullPath(Filename)
    End Function

    ' Versucht, aus einem Dateititel einer Musikdatei den Namen des Interpreten herauszufinden (Beispiel: "C:\Musik\Yocasa's Boys - Yeanui.mp3" --> "Yocasa's Boys"
    ' ACHTUNG: Der Versuch gelingt nicht in jedem Fall korrekt!
    Public Function GetArtistFromFileTitle(ByVal FileTitle As String) As String
        Dim Artist As String
        Dim Pos As Integer

        If FileTitle = "" Then Return "" ' eine Prüfung, ob die Datei existiert, ist nicht nötig
        Artist = FileTitle

        Pos = InStr(2, Artist, "-") ' 2, falls "-" Teil des Interpretennamen
        If Pos = 0 Then Pos = InStr(2, Artist, Chr(150)) ' Chr(150) = Langer Bindestrich
        If Pos > 0 Then
            If InStr(Pos + 1, Artist, "-") > 0 Then
                If IsNumeric(Mid(Artist, 1, Pos - 1)) Then Artist = Mid(Artist, Pos + 1, InStr(Pos + 1, Artist, "-") - Pos - 1)
            Else
                Artist = Mid(Artist, 1, Pos - 1)
            End If
        Else
            If InStr(1, Artist, ".") > 0 Then Artist = Mid(Artist, 1, InStr(1, Artist, ".") - 1)
        End If
        'Pos = InStr(2, Artist, "ft.", vbTextCompare) ' 2, falls "ft." zum Interpretennamen gehören sollte
        'If Pos = 0 Then Pos = InStr(2, Artist, "feat.", vbTextCompare) ' "
        'If Pos > 0 Then
        '    Artist = Left(Artist, Pos - 1)
        'End If
        Artist = Replace(Artist, "_", " ") ' Ersetzen von Unterstrichen durch Leerzeichen
        Return Trim(Artist) ' Hier zum 2. Mal Leerstellen entfernen
    End Function

    ' Versucht, aus einem Dateititel einer Musikdatei den Musiktitel herauszufinden (Beispiel: "C:\Musik\Yocasa's Boys - Yeanui.mp3" --> "Yeanui"
    ' ACHTUNG: Der Versuch gelingt nicht in jedem Fall korrekt!
    Public Function GetSongtitleFromFileTitle(ByVal FileTitle As String) As String
        Dim SongName As String
        Dim Pos As Integer

        If FileTitle = "" Then Return "" ' eine Prüfung, ob die Datei existiert, ist nicht nötig
        SongName = FileTitle

        Pos = InStr(2, SongName, "-") ' 2, falls "-" Teil des Interpretennamen
        If Pos = 0 Then Pos = InStr(2, SongName, Chr(150)) ' Chr(150) = Langer Bindestrich
        If Pos > 0 Then SongName = Mid(SongName, Pos + 1)

        Pos = InStr(1, SongName, "-")
        If Pos > 0 Then SongName = Mid(SongName, Pos + 1)

        SongName = Replace(SongName, "_", " ") ' Ersetzen von Unterstrichen durch Leerzeichen
        Return Trim(SongName) ' Hier zum 2. Mal Leerstellen entfernen
    End Function

    ' Konvertiert eine Zeichenfolge in einen String ohne führendes bzw. abschließendes "
    Public Function ConvertToRealString(ByVal Value As String) As String
        If Value.StartsWith("""") Then Value = Mid(Value, 2)
        If Value.EndsWith("""") Then Value = Left(Value, Value.Length - 1)
        Return Value
    End Function

    ' Gibt den verfügbaren Speicherplatz auf einem Laufwerk zurück.
    ' Angabe erfolgt in Bytes.
    ' Bsp. für Drive: "C" oder "C:"  oder "C:\" oder "c:\"
    Public Function GetFreeDriveSpace(ByVal Drive As String) As Long
        If Drive = "" Then
            Err.Raise(513, "GetFreeDriveSpace()", "Invalid Drive: " + Drive)
            Return IsVoid
        End If
        If Drive.Length = 1 Then Drive += ":"
        If Drive.Length = 2 Then Drive += "\"
        Drive = Drive.ToUpper

        Dim di As IO.DriveInfo
        For Each di In IO.DriveInfo.GetDrives()
            If di.Name.ToUpper = Drive Then
                Return di.AvailableFreeSpace
            End If
        Next
        Return IsVoid
    End Function

    ' Ist eine Datei/ein Ordner auf einem permanentem Datenträger gespeichert?
    ' Bzw. ist ein bestimmtes Laufwerk permanent?
    ' Die Datei/der Ordner muss nicht existieren; ausschlaggend ist nur das Laufwerk
    ' Nicht permanent sind CDs/DVDs, Flash-Memory (MP3-Player, USB-Sticks), Netzwerklaufwerke, RAM-Laufwerke...
    Public Function IsPermanent(ByVal Path As String) As Boolean
        If Len(Path) < 3 Then Return False

        Dim Drive As String
        Drive = Left(Path, 3) ' Laufwerk extrahieren
        Dim di As IO.DriveInfo
        For Each di In IO.DriveInfo.GetDrives()
            If di.Name.ToUpper = Drive Then
                Return CBool(di.DriveType = IO.DriveType.Fixed)
            End If
        Next
        Return False
    End Function
End Module

