﻿Public Class clsLYRDB
    '#######################################
    '#         clsLYRDB:                   #
    '# VB 2008 Wrapper for LYRDB's Lyrics  #
    '#       Database Service              #
    '#                                     #
    '#     class and service by            #
    '#    Flavio González Vázquez          #
    '#          flavio@ya.com              #
    '#                                     #
    '#    update from VB 6 to VB 2008 by   #
    '#    Christoph Konnertz               #
    '#        ckonnertz@aol.com            #
    '#                                     #
    '#                                     #
    '#    == Function credits ==           #
    '#    --> StringCount: Edward Moth     #
    '#    --> CrLfTrim (modified):         #
    '#                     Dane Jones      #
    '#                                     #
    '#    NOTICE:                          #
    '#    this version does not support    #
    '#      lyrics submissions!            #
    '#                                     #
    '#######################################

    '-- LYRDB Server Checking
    Private Const FLAG_ICC_FORCE_CONNECTION = &H1
    Private Declare Function InternetCheckConnection Lib "wininet.dll" Alias "InternetCheckConnectionA" (ByVal lpszUrl As String, ByVal dwFlags As Long, ByVal dwReserved As Long) As Long

    '-- URL Constants
    Private Const LYRDB_HELLO = "www.lyrdb.com"
    Private Const LYRDB_LOOKUP = "http://webservices.lyrdb.com/lookup.php"
    Private Const LYRDB_GETLYR = "http://www.lyrdb.com/getlyr.php"

    '-- Last lookup data
    Private Structure sLyrdbItem
        Dim sID As String
        Dim sTrackname As String
        Dim sArtist As String
    End Structure

    Private lItem() As sLyrdbItem

    Private mAgent As String

    Public Sub New()
        mAgent = vbNullString
    End Sub

    Public ReadOnly Property ServiceAvailable() As Boolean
        Get
            Try
                If My.Computer.Network.Ping(LYRDB_HELLO, 500) Then ' InternetCheckConnection(LYRDB_HELLO, FLAG_ICC_FORCE_CONNECTION, 0&) = 0
                    Return True
                Else
                    Return False
                End If
            Catch ex As Exception
                Return False
            End Try
        End Get
    End Property

    Public Function GetLyrdbHello() As String
        Dim buff As String = ""
        GetHTTPFile(LYRDB_HELLO, buff)
        Return Trim(buff)
    End Function

    ' Access the LYRDB.com database an searches a given query (sSearchStr).
    ' Returns the number of results. 0=none.
    Public Function LyrdbLookup(ByVal sField As String, ByVal sSearchStr As String, ByRef LastError As String) As Byte
        Dim results() As String, sRet() As String
        Dim buff As String = ""
        Dim n As Integer, pos As Integer

        sSearchStr = UrlEncode(sSearchStr) ' necessary substitutions
        GetHTTPFile(LYRDB_LOOKUP & "?q=" & sSearchStr & "&for=" & sField & "&agent=" & mAgent, buff)

        pos = InStr(buff, "<!--")
        If pos > 0 Then buff = Left(buff, pos - 1) ' removes hmtl tag at the end of the document
        buff = Trim(buff)

        If Left(buff, 6) = "error:" Then
            LastError = buff
            Return 0
        Else
            results = Split(buff, vbCrLf)

            If results(0) = "" Then
                Erase lItem
                Return 0
            Else
                ReDim lItem(0 To UBound(results))

                For n = 0 To UBound(results)
                    sRet = Split(results(n), "\")
                    lItem(n).sID = sRet(0)
                    If UBound(sRet) >= 1 Then lItem(n).sTrackname = sRet(1)
                    If UBound(sRet) >= 2 Then lItem(n).sArtist = sRet(2)
                Next

                Return UBound(results) + 1
            End If
        End If
    End Function

    Public Function GetLyrdbItem(ByVal sID As String) As String
        Dim buff As String = ""
        GetHTTPFile(LYRDB_GETLYR & "?q=" & sID, buff)

        buff = CrLfTrim(buff)

        If StringCount(buff, vbCrLf) = 0 Then
            Return ""
        Else
            Return buff
        End If
    End Function

    Private Function StringCount(ByVal Expression As String, ByVal item As String) As Integer
        Dim lPosition As Integer
        Dim lCount As Integer

        Do
            lPosition = InStr(lPosition + 1, Expression, item)
            If lPosition > 0 Then
                lCount = lCount + 1
            End If
        Loop While lPosition > 0
        Return lCount
    End Function

    Private Function CrLfTrim(ByVal StringToTrim As String) As String
        Dim chrToCheck As String
        Dim I As Byte

        For I = 1 To 2
            chrToCheck = Mid(StringToTrim, 1, 1)

            Do While chrToCheck = vbCr Or chrToCheck = vbLf Or chrToCheck = vbTab Or chrToCheck = " "
                StringToTrim = Mid(StringToTrim, 2)
                chrToCheck = Mid(StringToTrim, 1, 1)
            Loop
            StringToTrim = StrReverse(StringToTrim)
        Next I

        CrLfTrim = StringToTrim
    End Function

    Public ReadOnly Property Count() As Integer
        Get
            Try
                Count = UBound(lItem) + 1
                If Err.Number Then Count = 0
            Catch ex As Exception
                ' Do nothing.
            End Try
        End Get
    End Property

    Public ReadOnly Property LyrdbID(ByVal Index As Integer) As String
        Get
            Return lItem(Index).sID
        End Get
    End Property

    Public ReadOnly Property LyrdbTrackname(ByVal Index As Integer) As String
        Get
            Return lItem(Index).sTrackname
        End Get
    End Property

    Public ReadOnly Property LyrdbArtist(ByVal Index As Integer) As String
        Get
            Return lItem(Index).sArtist
        End Get
    End Property

    Public Property Agent() As String
        Get
            Return mAgent
        End Get
        Set(ByVal value As String)
            mAgent = value
        End Set
    End Property

    ' Downloads a existing html file from a known URL.
    ' The downlaoded file will be stored as one single String in sBuffer. vbCrLf will be used for separating lines.
    ' Returns true, if there wasnt any error. (This does not mean the the html file wasnt empty!)
    Function GetHTTPFile(ByVal URL As String, ByRef sBuffer As String) As Boolean
        If URL = "" Then Return False

        Try
            Dim req As System.Net.HttpWebRequest
            Dim resp As System.Net.HttpWebResponse
            Dim sr As IO.StreamReader
            Dim pos As Integer

            req = System.Net.HttpWebRequest.Create(URL)
            resp = req.GetResponse()
            sr = New IO.StreamReader(resp.GetResponseStream())

            If req.HaveResponse = False Or resp.ContentLength = 0 Then
                Return False
            End If

            While sr.EndOfStream = False
                sBuffer += sr.ReadLine() + vbCrLf
            End While
            sBuffer = Left(sBuffer, sBuffer.Length - 1)

            Return True
        Catch ex As Exception
            Return False
        End Try
    End Function

    ' Converts a given String into a valid URL
    Private Function UrlEncode(ByVal strBefore As String) As String
        Dim strAfter As String = ""
        Dim intLoop As Integer

        strBefore = Replace(strBefore, "'", "’")

        If Len(strBefore) > 0 Then
            For intLoop = 1 To Len(strBefore)
                Select Case Asc(Mid(strBefore, intLoop, 1))
                    Case 48 To 57, 65 To 90, 97 To 122, 46, 45, 95, 42  '0-9, A-Z, a-z . - _ *
                        strAfter = strAfter & Mid(strBefore, intLoop, 1)
                    Case 32
                        strAfter = strAfter & "+"
                    Case Else
                        strAfter = strAfter & "%" & Right("0" & Hex(Asc(Mid(strBefore, intLoop, 1))), 2)
                End Select
            Next
        End If

        Return strAfter.ToLower
    End Function
End Class
